=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed Jan. 2009 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:  body_Lib6G6.rb
# Original Date	:  03 Dec 2008 - version 3.0
# Type			:  Script library part of the LibFredo6 shared libraries
# Description	:  Contains some standalone generic geometric methods
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module G6

@retina_factor = Traductor.retina_factor

#---------------------------------------------------------------------------------------------------------------------------------
# Grouponents, that is: Group or Component Instances
#---------------------------------------------------------------------------------------------------------------------------------

#GROUPONENT: Check is an entity is a Group or a component instance
def G6.is_grouponent?(e)
	e.instance_of?(Sketchup::ComponentInstance) || e.instance_of?(Sketchup::Group)
end

#GROUPONENT: Determine the definition of of component or group
def G6.grouponent_definition(e)
	return e.definition if e.instance_of?(Sketchup::ComponentInstance)
	return e.entities.parent if e.instance_of?(Sketchup::Group)
	Sketchup.active_model
end

#GROUPONENT: Check if a grouponent is unique
def G6.grouponent_unique?(e)
	if e.instance_of?(Sketchup::Group) 
		cdef = e.entities.parent
	elsif e.instance_of?(Sketchup::ComponentInstance)
		cdef = e.definition
	else
		return true
	end	
	(cdef) ? (cdef.count_instances == 1 && cdef.instances.include?(e)) : true
end

#GROUPONENT: Make a grouponent is unique
def G6.grouponent_make_unique(e)
	if e.instance_of?(Sketchup::Group) 
		entities = e.entities
		cdef = entities.parent
		return entities if !cdef.instance_of?(Sketchup::ComponentDefinition) || cdef.count_instances == 1
		entities.add_group.erase!
		return entities
	elsif e.instance_of?(Sketchup::ComponentInstance)
		e.make_unique
		return e.definition.entities
	end
	Sketchup.active_model.active_entities 
end

#Check if a group is unique
def G6.is_group_unique?(g)
	return true unless g.instance_of?(Sketchup::ComponentDefinition) || g.instance_of?(Sketchup::Group)
	definition = (g.instance_of?(Sketchup::Group)) ? g.entities.parent : g.definition
	return true unless definition
	definition.count_instances == 1
end

#GROUPONENT: Return the entities, either for a Group or for a component
def G6.grouponent_entities(entity)
	return Sketchup.active_model.active_entities if !entity || entity == Sketchup.active_model
	(entity.class == Sketchup::ComponentInstance) ? entity.definition.entities : entity.entities
end

#GROUPONENT: Return the map of object id for all entities
#This is useful to retrieve the applicable entity object after a group or component has been made unique
def G6.grouponent_map_of_entities_object_id(g)
	return nil unless g.instance_of?(Sketchup::ComponentDefinition) || g.instance_of?(Sketchup::Group) 
	hmap = {}
	g.entities.each_with_index { |e, i| hmap[e.object_id] = i }
	hmap
end

# GROUPONENT: Compute the box around a component with offset
# Return the segment lines (list of pairs) to be drawn by view.draw GL_LINES
def G6.grouponent_box_lines(view, compo, tr=nil, pix=nil)
	pix = 1 unless pix
	tr = Geom::Transformation.new unless tr
	bb = G6.grouponent_definition(compo).bounds
	pts = [0, 1, 3, 2, 4, 5, 7, 6].collect { |i| tr * bb.corner(i) }
	ptsbox = pts.clone
	
	if pix > 0
		ptmid = Geom.linear_combination 0.5, pts[0], 0.5, pts[-1]
		size = view.pixels_to_model pix, ptmid
		
		normal = pts[0].vector_to pts[4]
		normal = (pts[0].vector_to pts[1]) * (pts[0].vector_to pts[2]) unless normal.valid?
		return [] unless normal.valid?
		[0, 1, 2, 3].each { |i| ptsbox[i] = ptsbox[i].offset normal, -size }
		[4, 5, 6, 7].each { |i| ptsbox[i] = ptsbox[i].offset normal, size }

		normal = pts[0].vector_to pts[1]
		normal = (pts[0].vector_to pts[3]) * (pts[0].vector_to pts[4]) unless normal.valid?
		return [] unless normal.valid?
		[0, 3, 4, 7].each { |i| ptsbox[i] = ptsbox[i].offset normal, -size }
		[1, 2, 5, 6].each { |i| ptsbox[i] = ptsbox[i].offset normal, size }

		normal = pts[0].vector_to pts[3]
		normal = (pts[0].vector_to pts[4]) * (pts[0].vector_to pts[1]) unless normal.valid?
		return [] unless normal.valid?
		[0, 1, 4, 5].each { |i| ptsbox[i] = ptsbox[i].offset normal, -size }
		[2, 3, 6, 7].each { |i| ptsbox[i] = ptsbox[i].offset normal, size }
	end
	
	[0, 1, 1, 2, 2, 3, 3, 0, 0, 4, 4, 5, 5, 6, 6, 7, 7, 4, 1, 5, 2, 6, 3, 7].collect { |i| ptsbox[i] } 
end

#GROUPONENT: Compute the 3D coordinates of the corners of the Bounding Box
def G6.grouponent_corners(comp, tr)
	bbox = G6.grouponent_definition(comp).bounds
	flat = (bbox.corner(0) == bbox.corner(4))
	corners3d = []
	range = (flat) ? 0..3 : 0..7
	for i in range
		corners3d.push tr * bbox.corner(i)
	end	
	corners3d
end

#GROUPONENT: Compute the full name of a container
def G6.grouponent_name(g)
	if g.class == Sketchup::Group
		name = "#{T6[:T_TXT_GROUP]} #{g.name}"
	elsif g.class == Sketchup::ComponentInstance
		cdef = g.definition
		if (g.name.empty?)
			imax = cdef.instances.length.to_s.length
			i = cdef.instances.rindex(g)
			inst_name = sprintf("%0#{imax}d", i+1)
		else
			inst_name = g.name
		end
		name = g.definition.name + " [#{inst_name}]"
	elsif g == nil || g == Sketchup.active_model
		name = 'Top Active Model'
	else
		name = 'unknown'
	end
	name
end

#GROUPONENT: Find top-level transformations for grouponents given by their id
#To be called only with first argument.
#Complement the value of the hash array
def G6.grouponent_find_transformations(hgrouponent_id, entities=nil, t=nil, n=nil)
	unless entities
		entities = Sketchup.active_model.active_entities
		n = hgrouponent_id.length
		t = Geom::Transformation.new
	end	
	
	lscomp = entities.grep(Sketchup::ComponentInstance) + entities.grep(Sketchup::Group)
	lscomp.each do |comp|
		comp_id = comp.entityID
		tnext = t * comp.transformation
		if hgrouponent_id.has_key?(comp_id)
			hgrouponent_id[comp_id] = tnext
			n -= 1
			return if n == 0
		elsif comp.instance_of?(Sketchup::Group)
			grouponent_find_transformations hgrouponent_id, comp.entities, tnext, n
		elsif comp.instance_of?(Sketchup::ComponentInstance)
			grouponent_find_transformations hgrouponent_id, comp.definition.entities, tnext, n
		end	
	end	
end

#GROUPONENT: Explode a group or component and return the drawing elements
def G6.grouponent_explode(g)
	g.explode.find_all { |e| e.valid? && e.is_a?(Sketchup::Drawingelement) }
end

def G6.transformation_is_mirrored?(t)
	vecx = t * X_AXIS
	vecy = t * Y_AXIS
	vecz = t * Z_AXIS
	normal = vecx * vecy
	(normal % vecz < 0)
end

#---------------------------------------------------------------------------------------------------------------------------------
# Current Editing context in Open components
#---------------------------------------------------------------------------------------------------------------------------------

#Return the Component or Group instance currently open (or nil if at top level)
def G6.which_instance_opened
	model = Sketchup.active_model
	ee = model.active_entities
	mm = ee.parent
	
	#Current context is at the toplevel
	return nil unless mm.class == Sketchup::ComponentDefinition
	
	#Getting the instances and finding the one that has a transformation equal to Identity
	lsti = mm.instances
	return lsti[0] if lsti.length == 1
	lsti.each do |instance|
		center = instance.bounds.center
		return instance if instance.transformation * center == center
	end
	nil
end

#Find the component Instances at top level of the model that use a given Component definition
def G6.find_top_comp_where_used(cdef, lst_top=nil)
	lst_top = [] unless lst_top
	return lst_top unless cdef.instance_of?(Sketchup::ComponentDefinition)
	cdef.instances.each do |comp| 
		parent = comp.parent
		if parent.instance_of? Sketchup::Model
			lst_top.push comp unless lst_top.include?(comp)
		elsif parent.instance_of? Sketchup::ComponentDefinition
			find_top_comp_where_used parent, lst_top
		end	
	end
	lst_top
end


#determine if a component is a dynamic component
def G6.is_dynamic_component?(comp)
	(comp.attribute_dictionary "dynamic_attributes") ? true : false
end

#---------------------------------------------------------------------------------------------------------------------------------
# Geometry Operations (to account for SU7 features)
#---------------------------------------------------------------------------------------------------------------------------------

#Map the model.start_operation method, managing the optional optimization last parameter in SU7
def G6.start_operation(model, title, speedup=false, *args)
	(SU_MAJOR_VERSION < 7) ? model.start_operation(title) : model.start_operation(title, speedup, *args)
end

def G6.continue_operation(model, title, speedup=false, *args)
	model.start_operation(title, speedup, *args) if (SU_MAJOR_VERSION >= 7)
end

#Compute the wireframe for a list of entities
#return a flat list of pairs of points to be usually drawn by view.draw GL_LINES
def G6.wireframe_entities(entities, hsh_edgeID=nil)
	G6._aux_wireframe_entities entities, Geom::Transformation.new, {}, [], 'model', hsh_edgeID
end

def G6._aux_wireframe_entities(entities, t, hedges, lcomp, parent, hsh_edgeID=nil)
	return [] unless entities
	entities.each do |e|
		if e.class == Sketchup::Edge
			next if hedges[e.to_s] == parent
			hedges[e.to_s] = parent
			lcomp.push t * e.start.position, t * e.end.position
			if hsh_edgeID
				hsh_edgeID[e.entityID] = true
				hsh_edgeID[e.start.entityID] = true
				hsh_edgeID[e.end.entityID] = true
			end	
		elsif e.class == Sketchup::Face
			e.edges.each do |edge|
				next if hedges[edge.to_s] == parent
				hedges[edge.to_s] = parent
				lcomp.push t * edge.start.position, t * edge.end.position
				if hsh_edgeID
					hsh_edgeID[edge.entityID] = true
					hsh_edgeID[edge.start.entityID] = true
					hsh_edgeID[edge.end.entityID] = true
				end	
			end	
		elsif e.class == Sketchup::Group
			G6._aux_wireframe_entities e.entities, t * e.transformation, hedges, lcomp, e, hsh_edgeID	
		elsif e.class == Sketchup::ComponentInstance
			G6._aux_wireframe_entities e.definition.entities, t * e.transformation, hedges, lcomp, e, hsh_edgeID	
		end	
	end
	lcomp	
end

#Transform a vector. The method accounts for non orthogonal axes, as often the case when shearing was applied
def G6.transform_vector(vector, t)
	return vector unless vector && vector.valid? && t
	axdt = vector.axes.collect { |v| t * v }
	vec = axdt[0] * axdt[1]
	vec.length = vector.length
	vec
end

#Project a vector to a plane
def G6.project_vector_to_plane(vec, plane)
	return vec unless vec.valid?
	pt1 = ORIGIN.project_to_plane plane
	pt2 = ORIGIN.offset(vec).project_to_plane plane
	pt1.vector_to(pt2)
end

#General average based on distance measure
def G6.multi_average(lval, ldist)
	#simple cases
	return nil if ldist.empty?
	n = ldist.length - 1
	if n < 0
		return nil
	elsif n == 0
		return lval[0]
	elsif n == 1
		d1, d2 = ldist
		z1, z2 = lval
		return((d1 * z2 + d2 * z1) / (d1 + d2))
	elsif n == 2
		d1, d2, d3 = ldist
		z1, z2, z3 = lval
		d12 = d1 * d2
		d13 = d1 * d3
		d23 = d2 * d3
		return((d12 * z3 + d13 * z2 + d23 * z1) / (d12 + d13 + d23))
	end

	#General case
	n = ldist.length - 1
	dsum = 0
	prod = []
	for i in 0..n
		prod[i] = 1
		for j in 0..n
			prod[i] *= ldist[j] unless i == j
		end
		dsum += prod[i]
	end	
	prod = prod.collect { |x| x / dsum }
	
	if lval[0].class == Array
		newval = []
		for k in 0..lval[0].length-1
			newval[k] = 0
			for i in 0..n
				newval[k] += lval[i][k] * prod[i]
			end	
		end	
	else
		newval = 0
		for i in 0..n
			newval += lval[i] * prod[i]
		end	
	end
	newval
end

#Check if a transformation is equivalent to the identity
def G6.transformation_identity?(t)
	(t * ORIGIN == ORIGIN && t * @@origin2 == @@origin2)
end
	
#---------------------------------------------------------------------------------------------------------------------------------
# OpenGL instructions
#---------------------------------------------------------------------------------------------------------------------------------

#Process the drawing of instructions
def G6.process_GL_instructions(view, t, lgl)
	return false unless lgl && lgl.length > 0
	view.drawing_color = 'black'
	view.line_stipple = ''
	view.line_width = 1
	
	lgl.each do |ll|
		code = ll[0]
		begin
			next unless ll[1]
			if code.class == String
				G6.view_draw_text view, t * ll[1], code
			elsif code.class == String && code == 'T'
				G6.view_draw_text view, t * ll[1], ll[2]
			else
				view.drawing_color = ll[2] if ll[2]
				view.line_width = ll[3] if ll[3]			
				view.line_stipple = ll[4] if ll[4]
				pts = ll[1].collect { |pt| t * pt }
				view.draw2d ll[0], pts
			end	
		rescue
			puts "problem with draw GL code = #{ll[0]} color = #{ll[2]}"
		end
	end	
	true
end

#OPEN GL: Instructions for a line with arrows
def G6.arrow_2D_instructions(pts, arrows_type, color, width, stipple, size=8)
	lgl = []
	lgl.push [GL_LINE_STRIP, pts, color, width, stipple]
	
	at_beg = at_end = at_beg_plain = at_end_plain = at_beg_open = at_end_open = false
	case arrows_type.to_s
	when /begin_*(.*)/
		at_beg = true
		specs = $1
		at_beg_plain = true if specs =~ /plain/
		at_beg_open = true if specs =~ /open/
	when /end_*(.*)/
		at_end = true
		specs = $1
		at_end_plain = true if specs =~ /plain/
		at_end_open = true if specs =~ /open/
	when /both_*(.*)/
		at_beg = at_end = true
		specs = $1
		at_beg_plain = at_end_plain = true if specs =~ /plain/
		at_beg_open = at_end_open = true if specs =~ /open/
	else
		return lgl
	end	
	
	size = 8 unless size
	size2 = size * 0.5
	#Arrow at beg point
	if at_beg && pts[0] != pts[1]
		pt0 = pts.first
		vec = pts[1].vector_to pt0
		vecn = vec * Z_AXIS
		ptm = pt0.offset vec, -size
		pt1 = ptm.offset vecn, size2
		pt2 = ptm.offset vecn, -size2
		lpt = [pt1, pt0, pt2]
		code = (at_beg_open && !at_beg_plain) ? GL_LINE_STRIP : GL_LINE_LOOP
		lgl.push [GL_POLYGON, lpt, color] if at_beg_plain
		lgl.push [code, lpt, color, width, '']
	end

	#Arrow at end point
	if at_end && pts[-1] != pts[-2]
		pt0 = pts.last
		vec = pts[-2].vector_to pt0
		vecn = vec * Z_AXIS
		ptm = pt0.offset vec, -size
		pt1 = ptm.offset vecn, size2
		pt2 = ptm.offset vecn, -size2
		lpt = [pt1, pt0, pt2]
		code = (at_end_open && !at_end_plain) ? GL_LINE_STRIP : GL_LINE_LOOP
		lgl.push [GL_POLYGON, lpt, color] if at_end_plain
		lgl.push [code, lpt, color, width, '']
	end
	
	lgl
end

def G6.instructions_cross_at_segment(pt1, pt2, color)
	pix = 6
	ptmid = Geom.linear_combination 0.5, pt1, 0.5, pt2
	xmid = ptmid.x
	ymid = ptmid.y
	pts = [Geom::Point3d.new(xmid-pix, ymid-pix), Geom::Point3d.new(xmid+pix, ymid+pix), 
	       Geom::Point3d.new(xmid-pix, ymid+pix), Geom::Point3d.new(xmid+pix, ymid-pix)]
	[GL_LINES, pts, color, 3, '']
end

def G6.instructions_forbidden_at_point(ptmid, pix=8)
	lgl = []
	pix = 8 unless pix
	pts = G6.pts_circle ptmid.x, ptmid.y, pix
	lgl.push [GL_POLYGON, pts, 'white']
	lgl.push [GL_LINE_LOOP, pts, 'red', 2, '']
	lgl.push [GL_LINE_STRIP, [pts[5], pts[11]], 'red', 2, '']
	lgl
end

#---------------------------------------------------------------------------------------------------------------------------------
# Text Drawing 2D
#---------------------------------------------------------------------------------------------------------------------------------

@@wid_filter = [["A-Z", 9], ["a-z", 7], [" ", 7], ["0-9", 8], [".,;:", 3]]
@@hgt_char = 16

def G6.text_true_length(text)
	return 0 unless text
	return text.length unless defined?(text.bytesize)
	nc = 0
	for i in 0..text.length-1
		s = text[i..i]
		nc += (s.bytesize > 2) ? 2 : 1
	end	
	nc	
end

#Compute the length in pixel of a single line text
def G6.simple_text_size(text)
	return [7, @@hgt_char] unless text
	nc = G6.text_true_length(text)
	wid = 0
	@@wid_filter.each do |ls|
		s, w = ls
		c = text.count(s)
		wid += w * c
		nc -= c
	end	
	wid += nc * 8
	wid -= 3 * text.count('il')
	wid -= 5 * text.count('I')
	[wid, @@hgt_char]
end

#Compute the length in pixel of a multi line text
def G6.text_size(text)
	return G6.simple_text_size(text) unless text
	ltx = text.split "\n"
	wtx = 0
	htx = 0
	ltx.each do |s|
		w, h = G6.simple_text_size(s)
		htx += h
		wtx = w if w > wtx
	end	
	[wtx, htx]
end

def G6.text_box_instructions(text, x, y, wid, hgt, justif=nil)
	return nil unless text
	justif = 'LT' unless justif
	ltx = text.split "\n"
	ntx = ltx.length
	ydec = 0
	xdec = 0
	
	#tStart Position for Y
	if justif =~ /B/i
		ybeg = hgt - @@hgt_char * ntx
	elsif justif =~ /H/i
		ybeg = (hgt - @@hgt_char * ntx) / 2 
	else
		ybeg = 0
	end	
	ybeg = y - ydec - ybeg
	
	#Coordinates of each line of text
	instructions = []
	ltx.each do |s|
		w, = G6.simple_text_size(s) if justif =~ /R|M/i
		if justif =~ /R/i
			xbeg = wid - w
		elsif justif =~ /M/i
			xbeg = (wid - w) / 2
		else
			xbeg = 0
		end	
		xbeg += x + xdec
		instructions.push [s, Geom::Point3d.new(xbeg, ybeg, 0)]
		ybeg -= @@hgt_char	
	end
	instructions
end

#Draw a small rectangle with the new text of the label
def G6.draw_rectangle_text(view, x, y, text, bk_color, fr_color, txdec=0, fr_width=2)
	return unless x
	hgttext = 16
	y = y - 25 - hgttext
	y = y + 40 if y < 0
	vpx = view.vpwidth - 32
	vpy = view.vpheight - 48
	len = G6.text_true_length(text) + txdec
	widtext = len * 7
	x = vpx - widtext - 6 if (x + widtext) > vpx
	y = vpy - hgttext - 3 if (y + hgttext) > vpy
	pts = G6.pts_rectangle x-5, y, widtext+4, hgttext + 4
	if @retina_factor != 1
		tretina = Geom::Transformation.scaling pts[0], @retina_factor
		pts = pts.collect { |pt| tretina * pt }
	end
	
	#Drawing the box
	view.drawing_color = bk_color
	view.draw2d GL_POLYGON, pts	
	view.line_width = fr_width * @retina_factor
	view.line_stipple = ''
	view.drawing_color = fr_color
	view.draw2d GL_LINE_LOOP, pts
	
	#drawing the text
	hmac = (RUN_ON_MAC) ? 4 : 0
	pt_text = Geom::Point3d.new(x, y+hmac)
	pt_text = tretina * pt_text if @retina_factor != 1
	view.draw_text pt_text, text
end

#Draw a small rectangle with the new text of the label
def G6.rectangle_text_instructions(text, x, y, bk_color, fr_color, txdec=0, fr_width=2, xdec=15)
	lgl = []
	return lgl unless x
	x += xdec
	hgttext = 16
	y = y - 25 - hgttext
	y = y + 40 if y < 0
	view = Sketchup.active_model.active_view
	vpx = view.vpwidth - 32
	vpy = view.vpheight - 48
	len = G6.text_true_length(text) + txdec
	widtext = len * 7
	x = vpx - widtext - 6 if (x + widtext) > vpx
	y = vpy - hgttext - 3 if (y + hgttext) > vpy
	pts = G6.pts_rectangle x-5, y, widtext+4, hgttext + 4
	
	#Drawing the box
	lgl.push [GL_POLYGON, pts, bk_color]
	lgl.push [GL_LINE_LOOP, pts, fr_color, fr_width, '']
	
	#drawing the text
	hmac = (RUN_ON_MAC) ? 4 : 0
	pt_text = Geom::Point3d.new(x, y+hmac)
	lgl.push ['T', pt_text, text]
	lgl
end

#Draw a small rectangle with the new text of the label
def G6.draw_rectangle_multi_text(view, x, y, text, hargs={})
	return unless x && text
	
	#Parameters
	bk_color = hargs.fetch :bk_color, 'lightyellow'
	fr_color = hargs.fetch :fr_color, 'green'
	fr_width = hargs.fetch :fr_width, 2
	dx = hargs.fetch :dx, 0
	dy = hargs.fetch :dy, 40
	xmargin = hargs.fetch :xmargin, 4
	ymargin = hargs.fetch :ymargin, 4
	justif = hargs.fetch :justif, 'LT'
	tr = hargs.fetch :tr, nil
	
	#Size of the text
	wtex, htex = G6.text_size text
	wbox = wtex + 2 * xmargin
	hbox = htex + 2 * ymargin
	
	#Positioning the box
	vpx = view.vpwidth - 32
	vpy = view.vpheight
	ybox = y - dy
	ybox = y + dy + hbox if ybox - hbox < 48 
	ybox = vpy if ybox > vpy
	xbox = x + dx
	xbox = vpx - wbox if xbox + wbox > vpx 
	xbox = 0 if xbox < 0 

	#Drawing the box
	pts = G6.pts_rectangle xbox, ybox - hbox, wbox, hbox
	if @retina_factor != 1
		tretina = Geom::Transformation.scaling pts[0], @retina_factor
		pts = pts.collect { |pt| tretina * pt }
	end
	
	view.drawing_color = bk_color
	view.draw2d GL_POLYGON, pts
	
	if fr_color && fr_width > 0
		view.line_width = fr_width * @retina_factor
		view.line_stipple = ''
		view.drawing_color = fr_color
		view.draw2d GL_LINE_LOOP, pts
	end
	
	#Drawing the text
	xtex = xbox + xmargin
	ytex = ybox - ymargin
	tr = Geom::Transformation.scaling(Geom::Point3d.new(xtex, ybox - hbox/2), 1, -1, 1) unless tr
	tr = tretina * tr if @retina_factor != 1
	lgl = G6.text_box_instructions text, xtex, ytex, wtex, htex, justif	
	G6.process_GL_instructions view, tr, lgl
end

#---------------------------------------------------------------------------------------------------------------------------------
# Inference Utilities
#---------------------------------------------------------------------------------------------------------------------------------

#Check whether inference can apply, given a hash table of Entity Ids
def G6.not_auto_inference?(ip, hsh_entID)
	return true unless hsh_entID
	v = ip.vertex
	e = ip.edge
	f = ip.face
	return false if (e && hsh_entID[e.entityID]) || (v && hsh_entID[v.entityID]) ||
	                (f && hsh_entID[f.entityID])
	true
end

#Determine if an inference is a true one (close to vertex) or a remote one
def G6.true_inference_vertex?(view, ip, x, y)
	vertex = ip.vertex
	return false unless vertex
	pt2d = view.screen_coords ip.transformation * vertex.position
	return false if (pt2d.x - x).abs > 10 || (pt2d.y - y).abs > 10
	true
end

#  Calculate the intersection points of a segment and a sphere
#  Credit to Paul Bourke (1992)- see http://local.wasp.uwa.edu.au/~pbourke/geometry/sphereline/
def G6.intersect_segment_sphere(pt1, pt2, center, radius)
	tolerance = 0.001.cm
   
	#Calculation
    dpx = pt2.x - pt1.x
    dpy = pt2.y - pt1.y
    dpz = pt2.z - pt1.z
   
	a = dpx * dpx + dpy * dpy + dpz * dpz
	b = 2 * (dpx * (pt1.x - center.x) + dpy * (pt1.y - center.y) + dpz * (pt1.z - center.z))
	c = center.x * center.x + center.y * center.y + center.z * center.z
	c += pt1.x * pt1.x + pt1.y * pt1.y + pt1.z * pt1.z
	c -= 2 * (center.x * pt1.x + center.y * pt1.y + center.z * pt1.z)
	c -= radius * radius
	bb4ac = b * b - 4 * a * c
   
	#NO solution
	return nil if (a.abs < tolerance || bb4ac < 0)

	#Computing the Intersection point
	sq = Math.sqrt(bb4ac)
    mu1 = (-b + sq) / (2 * a)
    mu2 = (-b - sq) / (2 * a)
	[Geom.linear_combination(1.0 - mu1, pt1, mu1, pt2), Geom.linear_combination(1.0 - mu2, pt1, mu2, pt2)]
end

#---------------------------------------------------------------------------------------------------------------------------------
# Stipple Properties
#---------------------------------------------------------------------------------------------------------------------------------

def G6.stipple_from_code(code)
	case code
	when 'D' ; '-'
	when 'U' ; '_'
	when 'P' ; '.'
	when 'A' ; '-.-'
	else ; ''
	end
end

#---------------------------------------------------------------------------------------------------------------------------------
# Edges Properties
#---------------------------------------------------------------------------------------------------------------------------------

#Dialog box for Dicing parameters - Return a list [dice, dice_format]
def G6.ask_edge_prop_filter(titletool, edge_prop)
	#Creating the dialog box
	hparams = {}
	title = ((titletool) ? titletool + ' - ' : '') + T6[:T_DLG_EdgeProp_Title]
	dlg = Traductor::DialogBox.new title
		
	enum_yesno = { 'Y' => T6[:T_DLG_YES], 'N' => T6[:T_DLG_NO] }
	ted = ' '
	dlg.field_enum "Plain", ted + T6[:T_DLG_EdgePlain], 'Y', enum_yesno
	dlg.field_enum "Soft", ted + T6[:T_DLG_EdgeSoft], 'Y', enum_yesno
	dlg.field_enum "Smooth", ted + T6[:T_DLG_EdgeSmooth], 'Y', enum_yesno
	dlg.field_enum "Hidden", ted + T6[:T_DLG_EdgeHidden], 'N', enum_yesno

	#Invoking the dialog box
	hparams["Plain"] = (edge_prop =~ /P/i) ? 'Y' : 'N'
	hparams["Soft"] = (edge_prop =~ /S/i) ? 'Y' : 'N'
	hparams["Smooth"] = (edge_prop =~ /M/i) ? 'Y' : 'N'
	hparams["Hidden"] = (edge_prop =~ /H/i) ? 'Y' : 'N'
	
	#Cancel dialog box
	return edge_prop unless dlg.show! hparams		
	
	#Transfering the parameters
	ep = []
	ep.push 'P' if hparams["Plain"] == 'Y'
	ep.push 'S' if hparams["Soft"] == 'Y'
	ep.push 'M' if hparams["Smooth"] == 'Y'
	ep.push 'H' if hparams["Hidden"] == 'Y'
	ep.join ';;'	
end

#check the filtering by edge property
def G6.edge_filter?(edge, filter)
	filter = 'P;;S;;M;;H' unless filter && filter.length > 0
	
	return false if edge.smooth? && !(filter =~ /M/i)
	return false if edge.soft? && !(filter =~ /S/i)
	return false if edge.hidden? && !(filter =~ /H/i)
	return false if !(edge.smooth? || edge.soft? || edge.hidden?) && !(filter =~ /P/i)
	true
end

#Check the filtering by edge property
def G6.edge_filter_text(filter)
	return T6[:T_DLG_EdgeAll] unless filter && filter.length > 0
	return T6[:T_DLG_EdgeAll] if filter =~ /P/i && filter =~ /S/i && filter =~ /M/i && filter =~ /H/i
	
	ls = []
	ls.push T6[:T_DLG_EdgePlain] if filter =~ /P/i
	ls.push T6[:T_DLG_EdgeSoft] if filter =~ /S/i
	ls.push T6[:T_DLG_EdgeSmooth] if filter =~ /M/i
	ls.push T6[:T_DLG_EdgeHidden] if filter =~ /H/i
	
	ls.join '+'
end

def G6.edge_prop_check_plain(prop)
	if prop == nil || prop == ''
		return 'P'
	elsif prop =~ /S|M|H/i
		lst = []
		lst.push 'S' if prop =~ /S/i
		lst.push 'M' if prop =~ /M/i
		lst.push 'H' if prop =~ /H/i
		return lst.join(';;')
	end	
	prop
end

#---------------------------------------------------------------------------------------------------------------------------------
# VECTOR: Geometry Utilities on Vectors
#---------------------------------------------------------------------------------------------------------------------------------

#VECTOR: Compute the oriented vector of an edge Start --> End
def G6.vector_edge(edge, t=nil)
	(t) ? (t * edge.start.position).vector_to(t * edge.end.position) : edge.start.position.vector_to(edge.end.position)
end

#VECTOR: Straight average of vectors
def G6.vector_straight_average(lvec)
	x = y = z = 0
	lvec.each { |v| v = v.normalize ; x += v.x ; y += v.y ; z += v.z }
	Geom::Vector3d.new(x, y, z).normalize
end

#VECTOR: Compute the average of a list of vectors - Return a normalized vector - Not really based on Choleski!!
def G6.vector_exact_average(lvec)
	n = lvec.length - 1
	lvec = lvec.collect { |v| v.normalize }
	case n
	when -1, 0
		return lvec[0]
		
	when 1	
		return Geom::Vector3d.linear_combination(0.5, lvec[0], 0.5, lvec[1]).normalize
	
	when 2
		lplane = lvec.collect { |v| [Geom::Point3d.new(v.x, v.y, v.z), v] }
		line = Geom.intersect_plane_plane lplane[0], lplane[1]
		return G6.vector_exact_average([lvec[0], lvec[2]]) unless line
		pt = Geom.intersect_line_plane line, lplane[2]
		return G6.vector_exact_average([lvec[0], lvec[1]]) unless pt
		return Geom::Vector3d.new(pt.x, pt.y, pt.z).normalize
		
	else
		G6.vector_straight_average lvec
	end
end

#VECTOR: Group the vectors by directions and returned a filtered list for further averaging
def G6.vector_grouping(lvec)
	n = lvec.length - 1
	return lvec if n <= 0
	ls = []
	for i in 0..n
		lvec_i = lvec[i]
		for j in i+1..n
			ls.push [lvec_i.angle_between(lvec[j]), i, j]
		end
	end	
	ls.sort! { |a, b| a.first <=> b.first }
	
	zero = 0.00001
	anglemax = ls[-1].first
	tol = anglemax * 0.5
	
	lvres = []
	for i in 0..n
		lvres[i] = [i]
	end	
	
	ls.each do |a|
		angle, i, j = a
		if angle < zero 
			lvres[j] = nil
		elsif angle <= tol && lvres[i] && lvres[j]
			lvres[i].push j
			lvres[j] = nil
		end
	end
	lvres = lvres.compact
	if anglemax < 0.5
		lvres = [G6.vector_straight_average(lvec)]
	else
		lvres = lvres.collect { |a| G6.vector_straight_average a.collect { |i| lvec[i] } }
	end
	lvres	
end

#VECTOR: Average vectors after grouping (for Joint Push Pull like usage)
def G6.vector_nice_average(lvec)
	G6.vector_exact_average(G6.vector_grouping(lvec))
end

#---------------------------------------------------------------------------------------------------------------------------------
# Geometry Utilities
#---------------------------------------------------------------------------------------------------------------------------------

#Compute an offset of a point by a vector even if vector is zero-length
def G6.robust_offset(pt, vec, d=nil)
	return pt.clone if !vec.valid? || d == 0
	(d) ? pt.offset(vec, d) : pt.offset(vec)
end

#Compute a plane from 3 points
def G6.plane_by_3_points(pt1, pt2, pt3)
	vec1 = pt1.vector_to pt2
	vec2 = pt1.vector_to pt3
	[pt1, vec1 * vec2]
end

#Convert 4-coordinates plane to [pt, normal]
def G6.plane_convert_coords(plane)
	return plane unless plane.class == Array && plane.length == 4
	normal = Geom::Vector3d.new *(plane[0..2])
	ptinter = Geom.intersect_line_plane [ORIGIN, normal], plane
	[ptinter, normal]
end

#calculate the normal to an edge of a face pointing toward the inside
def G6.normal_in_to_edge(edge, face)
	pt1 = edge.start.position
	pt2 = edge.end.position
	vec = face.normal * pt1.vector_to(pt2)
	vec.length = 1.0
	edge.reversed_in?(face) ? vec.reverse : vec
end

#Determine the intersection point between an edge and a line. return nil if no intersection
def G6.intersect_edge_line(edge, line)
	pt = Geom.intersect_line_line edge.line, line
	return nil unless pt
	ptbeg = edge.start.position
	return pt if pt == ptbeg
	ptend = edge.end.position
	return pt if pt == ptend
	(pt.vector_to(ptbeg) % pt.vector_to(ptend) < 0) ? pt : nil
end

#Determine if two edges and face for a convex corner at a vertex
def G6.convex_at_vertex(vertex, face, edge1, edge2, face2=nil)
	vother1 = edge1.other_vertex vertex
	vec1 = vertex.position.vector_to vother1.position
	vecin1 = G6.normal_in_to_edge edge1, face
	vecin2 = G6.normal_in_to_edge edge2, ((face2) ? face2 : face)
	((vec1 * vecin1) % (vecin1 * vecin2) >= 0) 
end

#Determine the bissector vector of two edges crossing at a vertex
#Origin of the returned vector is the vertex
def G6.bissector_edges_at_vertex(vertex, edge1, edge2, face)
	vother1 = edge1.other_vertex vertex
	vother2 = edge2.other_vertex vertex
	vec1 = vertex.position.vector_to vother1.position
	vec2 = vertex.position.vector_to vother2.position
	(vec1.parallel?(vec2)) ? vec1 * face.normal : Geom.linear_combination(0.5, vec1, 0.5, vec2) 
end

#Compute the other edge that is bordering the given face at a given vertex
def G6.other_edge_to_face_at_vertex(vertex, edge, face)
	vertex.edges.each do |e|
		return e if e != edge && (e.faces & [face] == [face])
	end
	nil
end

#Check if an edge should be smoothed by default. Smooth is not advised if angle at edge is too sharp
def G6.edge_would_smooth(edge, angle_break=nil)
	faces = edge.faces
	return false unless faces.length == 2
	f1, f2 = faces
	angle_break = 45.degrees unless angle_break
	angle = f1.normal.angle_between(f2.normal)
	(angle <= angle_break || angle >= Math::PI - angle_break)
end

#Calculate the average center of a list of point
def G6.segment_weighted_average(lseg_weight)
	return nil if lseg_weight.empty?
	x = y = z = sum_w = 0
	lseg_weight.each do |a|
		pt1, pt2, weight = a
		weight = 1 unless weight
		ptmid = Geom.linear_combination 0.5, pt1, 0.5, pt2
		x += ptmid.x * weight
		y += ptmid.y * weight
		z += ptmid.z * weight
		sum_w += weight
	end
	Geom::Point3d.new(x / sum_w, y / sum_w, z / sum_w)
end

#Calculate the average center of a list of point
def G6.straight_barycenter(lpt)
	return nil unless lpt && lpt.length > 0
	return lpt[0] if lpt.length == 1
	lpt = lpt[0..-2] if (lpt.first == lpt.last)
	x = y = z = 0
	lpt.each do |pt|
		x += pt.x
		y += pt.y
		z += pt.z
	end
	n = lpt.length
	Geom::Point3d.new x / n, y / n, z / n
end

#Calculate the average center of a contiguous curve
def G6.curve_barycenter(lpt)
	return nil unless lpt && lpt.length > 0
	return lpt[0] if lpt.length == 1
	lptmid = []
	dtot = 0.0
	x = y = z = 0
	for i in 1..lpt.length-1
		d = lpt[i].distance lpt[i-1]
		dtot += d
		ptmid = Geom.linear_combination 0.5, lpt[i], 0.5, lpt[i-1]
		x += ptmid.x * d
		y += ptmid.y * d
		z += ptmid.z * d
	end
	(dtot == 0) ? lpt[0] : Geom::Point3d.new(x / dtot, y / dtot, z / dtot)
end

#Compute the centroid and area of a face
def G6.face_centroid_area(face)
	lpt = face.outer_loop.vertices.collect { |v| v.position }
	G6.loop_centroid_area(lpt, face.normal)
end

#Compute the centroid and area of a face
def G6.loop_centroid_area(lpt, normal)
	axes = normal.axes
	tr_axe = Geom::Transformation.axes lpt[0], axes[0], axes[1], axes[2]
	tr_axe_inv = tr_axe.inverse
	lpt = lpt.collect { |pt| tr_axe_inv * pt }
	lpt.push lpt[0] unless lpt.first == lpt.last
	area = 0.0
	cx = cy = 0
	for i in 0..lpt.length-2
		pt1 = lpt[i]
		pt2 = lpt[i+1]
		a = pt1.x * pt2.y - pt2.x * pt1.y
		area += a
		cx += (pt1.x + pt2.x) * a
		cy += (pt1.y + pt2.y) * a
	end
	area *= 3
	[tr_axe * Geom::Point3d.new(cx / area, cy / area, 0), area]
end

#Check if a curve roughly fits a plane direction according to a given factor
#Return [quasi, plane, bary], where quasi is:
#   1 if perfectly plane
#  -1 if quasi plane
#   0 if not plane
def G6.curve_quasi_plane(pts, plane=nil, bary=nil, factor=0.1)
	plane = Geom.fit_plane_to_points pts unless plane
	bary = G6.curve_barycenter pts unless bary
	return [0, plane, bary] unless plane
	d_proj = pts.collect { |pt| pt.distance_to_plane(plane) }
	delta = d_proj.max - d_proj.min
	return [1, plane, bary] if delta < 0.00001
	
	n = d_proj.length
	m = 0
	sig = 0
	d_proj.each do |d| 
		m += d 
		sig += d * d
	end	
	m /= n
	sig = Math.sqrt(sig / n - m * m)
	
	dbary = pts.collect { |pt| bary.distance pt }
	[((sig < factor * dbary.max) ? -1 : 0), plane, bary]
end

#Determine the intersection point between a segment [ptbeg, ptend] and a line. return nil if no intersection
def G6.intersect_segment_line(ptbeg, ptend, line)
	return ptbeg if ptbeg.on_line?(line)
	return ptend if ptend.on_line?(line)
	pt = Geom.intersect_line_line [ptbeg, ptend], line
	(pt && pt.vector_to(ptbeg) % pt.vector_to(ptend) < 0) ? pt : nil
end

#Calculate the intersection of 2 segments
#Segments are a pair of points
def G6.intersect_segment_segment(seg1, seg2)
	ptinter = Geom.intersect_line_line seg1, seg2
	return nil unless ptinter
	within1 = (ptinter == seg1[0] || ptinter == seg1[1] || (ptinter.vector_to(seg1[0]) % ptinter.vector_to(seg1[1]) < 0))
	within2 = (ptinter == seg2[0] || ptinter == seg2[1] || (ptinter.vector_to(seg2[0]) % ptinter.vector_to(seg2[1]) < 0))
	(within1 && within2) ? ptinter : nil
end

#Compute the intersection of 2 segments in 3D space
#  --> Return nil if no intersection or [d, pt1, pt2] where
#         - d is an indication of proximity
#         - pt1 is the intersection point on first segment
#         - pt2 is the intersection point on second segment
def G6.segments_intersection3D(pt1beg, pt1end, pt2beg, pt2end, plane=nil)
	
	#Segments with null length or segments parallel
	vec1 = pt1beg.vector_to pt1end
	return nil unless vec1.valid?
	vec2 = pt2beg.vector_to pt2end
	return nil unless vec2.valid?
	return nil if vec1.parallel?(vec2)
	
	#True intersection
	d, pt1, pt2 = G6.proximity_segments pt1beg, pt1end, pt2beg, pt2end
	return [0, pt1, pt2] if d == 0
		
	#Plane of elevation
	plane = [ORIGIN, vec1 * vec2] unless plane
	ppt1beg, ppt1end, ppt2beg, ppt2end = [pt1beg, pt1end, pt2beg, pt2end].collect { |pt| pt.project_to_plane plane }
	
	d, = G6.proximity_segments ppt1beg, ppt1end, ppt2beg, ppt2end
	[d, pt1, pt2]
end

#Check if a set of points are coplanar
#  --> Return the plane or nil
def G6.points_coplanar?(pts)
	n = pts.length
	return nil if n < 3
	return Geom.fit_plane_to_points(pts) if n == 3
	
	n1 = [n / 2, 3].max
	n2 = [n - n1, 3].max
	plane1 = Geom.fit_plane_to_points *(pts[0..n1-1])
	plane2 = Geom.fit_plane_to_points *(pts[n-n2..n-1])
	normal1 = Geom::Vector3d.new *plane1[0..2]
	normal2 = Geom::Vector3d.new *plane2[0..2]
	(normal1.parallel?(normal2) && pts[0].on_plane?(plane2)) ? plane1 : nil
end

#Check if an edge is coplanar
def G6.edge_coplanar?(edge)
	dotmax = 0.9999999991	#based on thomthom advice
	faces = edge.faces
	(faces.length != 2) ? false : (faces.first.normal % faces.last.normal).abs > dotmax
end

def G6.lonely_vertex?(vx)
	edges = vx.edges
	return false if edges.length != 2
	(vx.position.on_line?([edges[0].other_vertex(vx).position, edges[1].other_vertex(vx).position]))
end

#Check the best plane for an edge given its neighbours
def G6.edge_alone_best_plane_normal(edge, pt, tr)
	pt = tr.inverse * pt
	
	#Closest vertex to the point
	vxbeg = edge.start
	vxend = edge.end
	dbeg = pt.distance vxbeg.position
	dend = pt.distance vxend.position
	vx0 = (dbeg < dend) ? vxbeg : vxend
	vec0 = vxbeg.position.vector_to vxend.position
	
	#Checking if there is an other edge at vertex
	edges = vx0.edges
	if edges.length != 2
		vx0 = (vx0 == vxbeg) ? vxend : vxbeg
		return G6.transform_vector(vec0.axes[1], tr) if vx0.edges.length != 2
	end

	#Trying to find a plane
	normal = nil
	while true
		edges = vx0.edges
		return G6.transform_vector(vec0.axes[1], tr) if edges.length != 2
		edge = edges.find { |e| e != edge }
		vec = edge.start.position.vector_to edge.end.position
		normal = vec0 * vec
		break if normal.valid?
		vx0 = edge.other_vertex vx0
	end	
	normal
end

#Check if a point pt is within the segment [pt1, pt2]
#  --> Return true or false
def G6.point_within_segment?(pt, pt1, pt2)
	return true if pt == pt1 || pt == pt2
	vec1 = pt.vector_to pt1
	vec2 = pt2.vector_to pt
	vec1.samedirection?(vec2)
end

#Intersection of a line and a triangle in 3D
#  --> Return the intersection point or nil
def G6.intersect_line_polygon(line, pts)
	plane = Geom.fit_plane_to_points(pts)
	origin = pts[0]
	vec1 = origin.vector_to pts[1]
	vec2 = origin.vector_to pts[2]
	normal = vec1 * vec2
	return nil unless normal.valid?
	ptinter = Geom.intersect_line_plane line, [origin, normal]
	return nil unless ptinter
	tr_axe = Geom::Transformation.axes(ptinter, *normal.axes).inverse
	pts_flat = pts.collect { |pt| tr_axe * pt }
	ptinter_flat = tr_axe * ptinter
	(Geom.point_in_polygon_2D(ptinter_flat, pts_flat, true)) ? ptinter : nil
end

#Check if a segment [pt1, pt2] crosses a given plane
#  --> Return the intersection point or nil
def G6.intersect_segment_plane(pt1, pt2, plane)
	return pt1 if pt1.on_plane?(plane)
	return pt2 if pt2.on_plane?(plane)
	pt = Geom.intersect_line_plane [pt1, pt2], plane
	return pt if pt == pt1 || pt == pt2
	(pt && (pt.vector_to(pt1) % pt.vector_to(pt2) < 0)) ? pt : nil
end

#Construct the list of points dividing a segment [pt1, pt2] into n subsegments
#  --> Return the list of points (so n+1 elements)
def G6.divide_segment(pt1, pt2, nbseg)
	pts = [pt1]
	ratio = 1.0 / nbseg
	for i in 1..nbseg-1
		f = i * ratio
		pts.push Geom.linear_combination(1-f, pt1, f, pt2)
	end
	pts.push pt2
	pts
end

#Check if a point in 2d (x, y) is close to a segment in 3d (pts)
def G6.close_to_segment_2d(view, x, y, pts, prox=8)
	n = pts.length - 1
	return nil if n == 0
	ptxy = Geom::Point3d.new x, y, 0
	for i in 0..n-1
		pt1_2d = view.screen_coords pts[i]
		pt2_2d = view.screen_coords pts[i+1]
		next if pt1_2d == pt2_2d
		ptproj = ptxy.project_to_line [pt1_2d, pt2_2d]
		next if ptxy.distance(ptproj) > prox
		return i if G6.point_within_segment?(ptproj, pt1_2d, pt2_2d)
	end	
	nil
end

#Check if 2 points are close on the screen
def G6.points_close_in_pixel?(view, pt1, pt2, pixels=5)
	vpt1 = view.screen_coords pt1
	vpt2 = view.screen_coords pt2
	vpt1.z = vpt2.z = 0
	(vpt1.distance(vpt2) <= pixels)
end

#BOX2D: Check if a polygon is inside or intersect a box in 2D
def G6.polygon_within_box2d?(pts2d, box2d)
	pts2d.each { |pt| return true if Geom.point_in_polygon_2D(pt, box2d, true) }
	box2d.each { |pt| return true if Geom.point_in_polygon_2D(pt, pts2d, true) }
	diag1 = [box2d[0], box2d[2]]
	diag2 = [box2d[1], box2d[3]]
	pts2d = pts2d + [pts2d.first] unless pts2d.first == pts2d.last
	for i in 0..pts2d.length-2
		line = [pts2d[i], pts2d[i+1]]
		return true if G6.intersect_segment_segment(line, diag1)
		return true if G6.intersect_segment_segment(line, diag2)
	end
	false	
end

#BOX2D: Check if a segment is inside or intersect a box in 2D
def G6.segment_within_box2d?(pt1, pt2, box2d)
	return true if Geom.point_in_polygon_2D(pt1, box2d, true)
	return true if Geom.point_in_polygon_2D(pt2, box2d, true)
	return true if G6.intersect_segment_segment([pt1, pt2], [box2d[0], box2d[2]])
	return true if G6.intersect_segment_segment([pt1, pt2], [box2d[1], box2d[3]])
	false	
end

#---------------------------------------------------------------------------------------------------------------------------------
# CLIPPING: Clipping on a rectangle
#---------------------------------------------------------------------------------------------------------------------------------

#CLIP: Clipping a segment in a rectangle in 2D
def G6.clip2d_rect_segment(seg, rect)
	pt1, pt2 = seg
	lnew = []
	[[0, 1], [1, 2], [2, 3], [3, 0]].each do |i, j|
		ptrect1 = rect[i]
		ptrect2 = rect[j]
		ptinter = Geom.intersect_line_line [ptrect1, ptrect2], seg
		next unless ptinter	&& G6.point_within_segment?(ptinter, ptrect1, ptrect2)
		next if ptinter == pt1 || ptinter == pt2
		if G6.point_within_segment?(ptinter, pt1, pt2)
			if lnew[0] == nil
				lnew[0] = [ptinter, i]
			elsif lnew[1] == nil
				lnew[1] = [ptinter, i]
			end	
		end
	end	
	lnew = lnew.reverse if lnew.length == 2 && pt1.distance(lnew[0][0]) >  pt1.distance(lnew[1][0])

	within1 = Geom.point_in_polygon_2D(pt1, rect, true)
	within2 = Geom.point_in_polygon_2D(pt2, rect, true)
	ptbeg = (within1) ? [pt1, nil] : ((lnew[0]) ? lnew[0] : lnew[1])
	ptend = (within2) ? [pt2, nil] : ((lnew[1]) ? lnew[1] : lnew[0])
	(ptbeg && ptbeg[0] && ptend && ptend[0]) ? [ptbeg, ptend] : nil
end

#CLIP: Clipping a triangle in a rectangle in 2D
def G6.clip2d_rect_triangle(triangle, rect)
	bbox_r = Geom::BoundingBox.new
	bbox_r.add rect
	bbox_t = Geom::BoundingBox.new
	bbox_t.add triangle
	bb = bbox_r.intersect(bbox_t)
	return nil if bb.empty?

	pt0, pt1, pt2 = triangle
	pt0.z = pt1.z = pt2.z = 0
	withins_q = rect.collect { |pt| (Geom.point_in_polygon_2D(pt, triangle, true)) ? pt : nil }
	withins_t = triangle.collect { |pt| Geom.point_in_polygon_2D(pt, rect, true) }
	
	#Case where triangle is either fully enclosed in box or box fully encloses triangle
	return rect if (withins_q.find_all { |a| a }).length == 4
	return triangle if (withins_t.find_all { |a| a }).length == 3
	
	#Case where there is at least an intersection
	links = []
	for i in 0..2
		j = (i == 2) ? 0 : i+1
		lk = G6.clip2d_rect_segment([triangle[i], triangle[j]], rect)
		links.push lk if lk
	end	
	
	lnew = []
	n = links.length-1
	for i in 0..n
		j = (i == n) ? 0 : i+1
		link_a = links[i] 
		lnew.push link_a[0][0], link_a[1][0]
		link_b = links[j] 
		ka = link_a[1][1]
		kb = link_b[0][1]
		next unless ka && kb && ka != kb
		
		#Checking if the rectangle corner is in the triangle
		pt = withins_q[ka]
		lnew.push pt if pt

		#Checking if adjacent rectangle corners are in the triangle
		k = (ka+1).modulo(4)
		pt = withins_q[k]
		if pt 
			lnew.push pt
			incr = +1
		else
			k = (ka-1).modulo(4)
			pt = withins_q[k]
			if pt
				lnew.push pt
				incr = -1
			else
				next
			end
		end	
		while true
			k = (k+incr).modulo(4)
			pt = withins_q[k]
			lnew.push pt if pt
			break if k == kb
		end	
	end
	
	lnew
end

#ALGO: Correction for a bug in view.screen_coords
def G6.check_screen_coords(view, pt_3d)
	return nil if view.camera.direction % view.camera.eye.vector_to(pt_3d) < 0
	pt = view.screen_coords(pt_3d)
	pt.z = 0
	pt
end

#---------------------------------------------------------------------------------------------------------------------------------
# UNITS: management of units
#---------------------------------------------------------------------------------------------------------------------------------

#UNITS: Return the symbol of the model unit
def G6.model_units
   current_unit = Sketchup.active_model.options["UnitsOptions"]["LengthUnit"]
   [:inch, :feet, :mm, :cm, :m][current_unit]
end

#UNITS: Check if model units are in Architectural mode
def G6.model_units_architectural?
  Sketchup.active_model.options["UnitsOptions"]["LengthUnit"] < 2
end

#UNITS: Check if model units are in Architectural mode
def G6.model_units_precision
  Sketchup.active_model.options["UnitsOptions"]["LengthPrecision"]
end

#UNITS: Format length including very small
def G6.format_length_general(len)
	archi = G6.model_units_architectural?
	leng = (archi) ? len : len.to_mm
	p = G6.model_units_precision
	if leng < (10000 / 10**p)
		n = 1
		n = 5 if leng < 0.0001
		n = 4 if leng < 0.001
		n = 3 if leng < 0.01
		n = 2 if leng < 0.1
		n = 1 if leng < 1.0
		n = 0 if leng > 10.0
		format = "%0.#{n}f"
		if archi
			text = "#{sprintf format, leng}\""
		else
			text = "#{sprintf format, leng} mm"
		end	
	else	
		text = Sketchup.format_length len
	end
	text
end

#---------------------------------------------------------------------------------------------------------------------------------
# PROXIMITY methods: points, segments, curves
#---------------------------------------------------------------------------------------------------------------------------------

#Compute the minimum distance and proximity points between two segments	
def G6.proximity_segments(pt1beg, pt1end, pt2beg, pt2end)
	ls = Geom.closest_points [pt1beg, pt1end], [pt2beg, pt2end]
	
	pt1 = G6.proximity_fit_point_within_segment ls[0], pt1beg, pt1end
	pt2 = G6.proximity_fit_point_within_segment ls[1], pt2beg, pt2end

	pt2 = pt1.project_to_line [pt2beg, pt2end]
	pt2 = G6.proximity_fit_point_within_segment pt2, pt2beg, pt2end

	pt1 = pt2.project_to_line [pt1beg, pt1end]
	pt1 = G6.proximity_fit_point_within_segment pt1, pt1beg, pt1end
	
	[pt1.distance(pt2), pt1, pt2]
end 

#Find the closest point within the segment, given a point on the SAME line
def G6.proximity_fit_point_within_segment(pt, ptbeg, ptend)
	return pt if pt == ptbeg || pt == ptend || pt.vector_to(ptbeg) % pt.vector_to(ptend) < 0
	(pt.vector_to(ptbeg) % ptbeg.vector_to(ptend) < 0) ? ptend : ptbeg
end

#Compute the minimum distance and proximity points between a point and a segment
def G6.proximity_point_segment(pt, ptbeg, ptend)
	pt2 = pt.project_to_line [ptbeg, ptend]
	pt2 = G6.proximity_fit_point_within_segment(pt2, ptbeg, ptend)
	[pt.distance(pt2), pt, pt2]
end

#Compute the minimum distance and proximity points between two curves	
def G6.proximity_curves(pts1, pts2)	
	return nil unless pts1 && pts2 && pts1.length > 0 && pts2.length > 0
	n1 = pts1.length - 1
	n2 = pts2.length - 1
	pts1_first = pts1[0]
	pts2_first = pts2[0]
	
	#Curves reduced to a single point
	if n1 == 0
		return [pts1_first.distance(pts2_first), 1, 1, pts1_first, pts2_first] if n2 == 0
		return G6.proximity_point_curve(pts1_first, pts2)
	end
	
	if n2 == 0
		res = G6.proximity_point_curve(pts2_first, pts1)
		return [res[0], res[2], res[1], res[4], res[3]]
	end
	
	#Compute the proximity parameters between curves
	d0 = pts1_first.distance pts2_first
	res = [d0, 1, 1, pts1_first, pts2_first]
	
	for i in 1..n1
		pt1beg = pts1[i-1]
		pt1end = pts1[i]
		for j in 1..n2
			pt2beg = pts2[j-1]
			pt2end = pts2[j]
			d, pt1, pt2 = G6.proximity_segments pt1beg, pt1end, pt2beg, pt2end
			if d < d0
				res = [d, i, j, pt1, pt2]
				return res if d == 0
				d0 = d
			end
		end
	end	
	res
end

#Compute the minimum distance and proximity points between a point and a curve
def G6.proximity_point_curve(pt, pts, d_to_beat=nil, iseg_beg=0)
	d0 = pt.distance pts.first
	res = [d0, 1+iseg_beg, pt, pts.first]
	d0 = d_to_beat if d_to_beat
	
	for i in 1..pts.length-1
		ptbeg = pts[i-1]
		ptend = pts[i]	
		ptproj = pt.project_to_line [ptbeg, ptend]		
		d = pt.distance(ptproj)
		next if d >= d0
		
		dbeg = ptproj.distance ptbeg
		dend = ptproj.distance ptend
		if dbeg + dend > ptbeg.distance(ptend) + 0.0001
			ptproj = (dbeg < dend) ? ptbeg : ptend
			d = pt.distance(ptproj)	
		end	
		if d < d0
			res = [d, i+iseg_beg, pt, ptproj]
			return res if d == 0
			d0 = d
		end
	end	
	res
end

#CONTOUR: Compute the minimum distance between a point and a curve
def G6.proximity_point_curve_by_circle(pt, circles, dmin=nil)
	lsorder = circles.collect do |circle|
		ptcenter, radius, pts, ibeg = circle
		[ptcenter.distance(pt) - radius, pts, ibeg]
	end
	lsorder.sort! { |a, b| a.first <=> b.first }
	
	resmin = nil
	lsorder.each do |a|
		d, pts, ibeg = a
		if dmin
			break if d > dmin
			res = G6.proximity_point_curve(pt, pts, dmin, ibeg)
			if res.first < dmin
				resmin = res 
				dmin = res.first
			end	
		else
			resmin = G6.proximity_point_curve(pt, pts, dmin, ibeg)
			dmin = resmin.first		
		end	
	end	
	resmin
end

#CONTOUR: Calculate the Boundary box information
def G6.contour_compute_proximity_circles(pts, nseg_circle=7)
	circles = []	
	
	#Computing the length of segments and total length
	dtot = 0
	dist = []
	for i in 0..pts.length-2
		dist[i] = pts[i].distance pts[i+1]
		dtot += dist[i]
	end
	
	dmax = dtot / nseg_circle
	d = 0
	ibeg = 0
	for i in 0..pts.length-2
		d += dist[i]
		if d > dmax 
			circles.push G6.contour_circle_info(pts[ibeg..i+1], ibeg)
			d = 0
			ibeg = i + 1
		end
	end	
	circles.push G6.contour_circle_info(pts[ibeg..-1], ibeg) if d > 0
	circles
end

#CONTOUR: Calculate the circle information for a portion of the contour
def G6.contour_circle_info(pts, ibeg)
	ptcenter = G6.curve_barycenter pts
	radius = 0
	pts.each do |pt|
		d = ptcenter.distance pt
		radius = d if d > radius
	end
	[ptcenter, radius * 1.001, pts, ibeg] 
end

#---------------------------------------------------------------------------------------------------------------------------------
# Mirror Transformation
#---------------------------------------------------------------------------------------------------------------------------------

#Compute the mirror of a list of points about a given plane
def G6.tr_mirror_about_plane(origin, normal)
	if normal.parallel?(Z_AXIS)
		trot = Geom::Transformation.new
	else	
		trot = Geom::Transformation.rotation origin, normal * Z_AXIS, normal.angle_between(Z_AXIS)
	end	
	ts = Geom::Transformation.scaling origin, 1, 1, -1
	trot.inverse * ts * trot
end

#---------------------------------------------------------------------------------------------------------------------------------
# SU Colors
#---------------------------------------------------------------------------------------------------------------------------------

@@sel_colors = nil
@@su_colors = nil

#Colors useful for edge selection (light colors)
def G6.color_selection(i)
	unless @@sel_colors
		@@sel_colors = ['orange', 'yellow', 'tomato', 'red', 'gold', 'lightgreen', 'coral', 'salmon', 
					    'orangered', 'sandybrown', 'greenyellow']
	end
	@@sel_colors[i.modulo(@@sel_colors.length)]
end

#Sketchup Colors
def G6.color_su(i)
	unless @@su_colors
		@@su_colors = Sketchup::Color.names.find_all do |n| 
			c = Sketchup::Color.new(n)
			g = c.red + c.blue + c.green
			g < 510 && g > 100 && g != 255
		end	
	end
	@@su_colors[i.modulo(@@su_colors.length)]
end

#Color just lighter than the edge color
def G6.color_edge_sel
	color_sel = Sketchup.active_model.rendering_options['ForegroundColor']
	dec = 80
	lc = [color_sel.red, color_sel.blue, color_sel.green].collect { |c| c + dec }
	Sketchup::Color.new *lc
end

#---------------------------------------------------------------------------------------------------------------------------------
# Edges Around a surface
#---------------------------------------------------------------------------------------------------------------------------------

#Determine all connected faces to the face (i.e. if bordering edge is soft or hidden)
#note: the recursive version seems to bugsplat on big number of faces. So I use an iterative version
def G6.face_neighbours(face, hsh_faces=nil)
	lface = [face]
	hsh_faces = {} unless hsh_faces
	
	while lface.length > 0
		f = lface.shift
		next if hsh_faces[f.entityID]
		hsh_faces[f.entityID] = f
		f.edges.each do |e|
			if e.soft? || e.smooth? || e.hidden?
				e.faces.each do |ff| 
					lface.push ff unless ff == f || hsh_faces[ff.entityID]
				end	
			end	
		end
	end	
	hsh_faces.values
end

#Calculate the contour of the surface (i.e. edges with only one face)
#Return as a Hash table, indexed by entityID of edges
def G6.edges_around_face(face, hsh_good_edges=nil, hsh_bad_edges=nil)
	#calculate the neighbour faces
	hsh_faces = {}
	hsh_good_edges = {} unless hsh_good_edges
	hsh_bad_edges = {} unless hsh_bad_edges
	G6.face_neighbours face, hsh_faces
	
	#Calculate the bordering edges
	hsh_good_edges = {}
	hsh_faces.each do |key, face|
		face.outer_loop.edges.each do |e|
			n = 0
			e.faces.each { |f| n += 1 if hsh_faces[f.entityID] }
			if (n == 1)
				hsh_good_edges[e.entityID] = e
			else
				hsh_bad_edges[e.entityID] = e
			end	
		end
	end	
	hsh_good_edges.values
end

#Compute the contours of the faces belonging to the session
def G6.contour_of_faces(hsh_or_lst_faces)
	return nil unless hsh_or_lst_faces
	
	#If a list of face is given, transform it into a hash array
	if hsh_or_lst_faces.class == Array
		hsh_faces = {}
		hsh_or_lst_faces.each { |f| hsh_faces[f.entityID] = f }
	elsif hsh_or_lst_faces.class == Hash
		hsh_faces = hsh_or_lst_faces
	else
		return nil
	end	
	
	#Computing the edges which have only one bordering face of the group of faces
	hsh_edges = {}	
	hsh_faces.each do |key, face|
		face.edges.each do |edge|
			next if hsh_edges[edge.entityID]
			n = 0
			edge.faces.each do |f|
				n += 1 if hsh_faces[f.entityID]
				break if n > 1
			end	
			hsh_edges[edge.entityID] = edge if n == 1
		end			
	end
	
	#Creating the contour as a list of lines
	contour = []
	hsh_edges.each do |key, edge|
		contour.push edge.start.position, edge.end.position
	end	
	contour
end

#Compute the angle between 2 edges at a given vertex
def G6.edges_angle_at_vertex(e1, e2, vertex)
	v1 = e1.other_vertex vertex
	v2 = e2.other_vertex vertex
	vec1 = vertex.position.vector_to(v1).normalize
	vec2 = vertex.position.vector_to(v2).normalize
	vec1.angle_between vec2
end

#Compute the normal to 2 edges
def G6.edges_normal(e1, e2)
	vec1 = e1.start.position.vector_to e1.end.position
	vec2 = e2.start.position.vector_to e2.end.position
	vec1 * vec2
end

#Transfer the properties from edge e1 to edge e2
def G6.edge_transfer_properties(e1, e2)
	e2.soft = e1.soft?
	e2.smooth = e1.smooth?
	e2.hidden = e1.hidden?
	e2.casts_shadows = e1.casts_shadows?
	e2.material = e1.material
	e2.layer = e1.layer
end

#Transfer the properties from edge e1 to edge e2
def G6.face_transfer_properties(f1, f2)
	f2.material = f1.material
	f2.back_material = f1.back_material
	f2.layer = f1.layer
end

#Transfer Faces to another entities
def G6.copy_faces_from(lfaces, entities, tr)
	#Sorting faces, with holes first
	faces_hole = []
	faces_plain = []
	lfaces.each do |face|
		if face.loops.length > 1
			faces_hole.push face
		else
			faces_plain.push face
		end
	end
	lfaces = faces_hole + faces_plain
	
	#Copying the faces
	new_faces = []
	lfaces.each do |face|
		face.loops.each_with_index do |loop, iloop|
			pts = loop.vertices.collect { |vx| tr * vx.position }
			begin
				f = entities.add_face pts
				if iloop > 0
					f.erase!
				else
					G6.face_transfer_properties(face, f)
					new_faces.push f
				end	
			rescue
			end
		end
	end

	new_faces
end

#Transfer Faces to another entities
def G6.copy_edges_from(ledges, entities, tr)	
	new_edges = []
	ledges.each do |edge|
		pt1 = tr * edge.start.position
		pt2 = tr * edge.end.position
		new_edge = entities.add_line pt1, pt2
		G6.edge_transfer_properties(edge, new_edge)
		new_edges.push new_edge
	end

	new_edges
end

#Transfer Faces to another entities
def G6.copy_clines_from(lclines, entities, tr)	
	new_clines = []
	lclines.each do |cline|
		pt1 = tr * cline.start
		pt2 = tr * cline.end
		new_cline = entities.add_cline pt1, pt2
		new_clines.push new_cline
	end

	new_clines
end

#Test if an edge is Plain
def G6.edge_plain?(e)
	!(e.smooth? || e.soft? || e.hidden?)
end

#Test is an edge is a diagonal
def G6.edge_is_diagonal?(edge)
	!edge.casts_shadows?
end

#Mark an edge is a diagonal
def G6.mark_diagonal(edge)
	edge.casts_shadows = !edge.casts_shadows?
end

#Find non-colinear vertices on a face
def G6.face_best_three_vertices(face)
	vertices = face.vertices
	v0 = vertices[0]
	v1 = vertices[1]
	vertices[2..-1].each do |vx|
		return [v0, v1, vx] unless vx.position.on_line?([v0.position, v1.position])
	end
	vertices[0..2]
end

#Calculate the list of edges in order from a sequence of vertices	
def G6.edges_from_vertices_in_sequence(lvx)
	ledges = []
	for i in 0..lvx.length-2
		edge = lvx[i].common_edge lvx[i+1]
		ledges.push edge if edge
	end	
	ledges
end

#Calculate the list of vertices in order from a sequence of edges	
def G6.vertices_from_edges_in_sequence(ledges, ptfirst=nil)
	edge0 = ledges[0]
	vx0 = (edge0.end.position == ptfirst) ? edge0.end : edge0.start 
	lvx = [vx0]
	ledges.each { |edge| lvx.push edge.other_vertex(lvx.last) }	
	lvx
end

#---------------------------------------------------------------------------------------------------------------------------------
# Viewport drawing utilities
#---------------------------------------------------------------------------------------------------------------------------------

#Draw a text in a view (with correction of Y for Mac)
@@is_mac = (RUBY_PLATFORM =~ /darwin/i)
def G6.view_draw_text(view, pt, text)
	if @@is_mac 
		pt = pt.clone
		pt.y += 3
	end
	view.draw_text pt, text
end

#Draw all selected edges
def G6.draw_lines_with_offset(view, lpt, color, width, stipple, pix=1)
	return if lpt.length == 0
	lpt = lpt.collect { |pt| G6.small_offset view, pt, pix }
	view.line_width = width
	view.line_stipple = stipple
	view.drawing_color = color
	view.draw GL_LINES, lpt
end

def G6.draw_gl_with_offset(view, gl_code, lpt, pix=1)
	return if lpt.length < 2
	lpt = lpt.collect { |pt| G6.small_offset view, pt, pix }
	view.draw gl_code, lpt
end

#Calculate the point slightly offset to cover the edges
def G6.small_offset(view, pt, pix=1)
	pt2d = view.screen_coords pt
	return pt if pt2d.x > 200000 || pt2d.y > 200000
	ray = view.pickray pt2d.x, pt2d.y
	vec = ray[1]
	size = view.pixels_to_model pix.abs, pt
	size = -size if pix < 0
	pt.offset vec, -size
end

#Compute the points of a square centered at x, y with side 2 * dim
def G6.pts_square(x, y, dim)
	pts = []
	pts.push Geom::Point3d.new(x-dim, y-dim)
	pts.push Geom::Point3d.new(x+dim, y-dim)
	pts.push Geom::Point3d.new(x+dim, y+dim)
	pts.push Geom::Point3d.new(x-dim, y+dim)
	pts
end

#Compute the points of a square centered at x, y with side 2 * dim
def G6.pts_rectangle(x, y, dimx, dimy)
	pts = []
	pts.push Geom::Point3d.new(x, y)
	pts.push Geom::Point3d.new(x+dimx, y)
	pts.push Geom::Point3d.new(x+dimx, y+dimy)
	pts.push Geom::Point3d.new(x, y+dimy)
	pts
end

#Compute the points of a square centered at x, y with side 2 * dim
def G6.pts_round_rectangle(x, y, dimx, dimy, radius, n=2)
	angle = -0.5 * Math::PI / (n+1)	
	pts = []
	
	center = Geom::Point3d.new(x+radius, y+radius)
	trot = Geom::Transformation.rotation center, Z_AXIS, angle
	pt0 = center.offset(Y_AXIS.reverse, radius)
	pts.push pt0
	for i in 1..n+1
		pt0 = trot * pt0
		pts.push pt0
	end
	
	center = Geom::Point3d.new(x+radius, y+dimy-radius)
	trot = Geom::Transformation.rotation center, Z_AXIS, angle
	pt0 = center.offset(X_AXIS.reverse, radius)
	pts.push pt0
	for i in 1..n+1
		pt0 = trot * pt0
		pts.push pt0
	end

	center = Geom::Point3d.new(x+dimx-radius, y+dimy-radius)
	trot = Geom::Transformation.rotation center, Z_AXIS, angle
	pt0 = center.offset(Y_AXIS, radius)
	pts.push pt0
	for i in 1..n+1
		pt0 = trot * pt0
		pts.push pt0
	end

	center = Geom::Point3d.new(x+dimx-radius, y+radius)
	trot = Geom::Transformation.rotation center, Z_AXIS, angle
	pt0 = center.offset(X_AXIS, radius)
	pts.push pt0
	for i in 1..n+1
		pt0 = trot * pt0
		pts.push pt0
	end

	pts
end

#Compute the points of a square centered at x, y with side 2 * dim
def G6.pts_triangle(x, y, dim=2, vdir=Y_AXIS)
	pts = []
	pts.push Geom::Point3d.new(x-dim, y-dim)
	pts.push Geom::Point3d.new(x+dim, y-dim)
	pts.push Geom::Point3d.new(x, y+dim)
	
	angle = Y_AXIS.angle_between vdir
	angle = -angle if (Y_AXIS * vdir) % Z_AXIS < 0
	return pts if angle == 0
	
	ptmid = Geom::Point3d.new x, y, 0
	t = Geom::Transformation.rotation ptmid, Z_AXIS, angle
	pts.collect { |pt| t * pt }
end

#Compute the points of a circle centered at x, y with radius
def G6.pts_circle(x, y, radius, n=12)
	pts = []
	angle = Math::PI * 2 / n
	for i in 0..n
		a = angle * i
		pts.push Geom::Point3d.new(x + radius * Math.sin(a), y + radius * Math.cos(a))
	end	
	pts
end

#Compute the points of a cross centered at x, y with half-length
#Return the 4 poinst in sequences for use by GL_LINES
def G6.pts_cross(x, y, half_length)
	pt1 = Geom::Point3d.new x - half_length, y, 0
	pt2 = Geom::Point3d.new x + half_length, y, 0
	pt3 = Geom::Point3d.new x, y + half_length, 0
	pt4 = Geom::Point3d.new x, y - half_length, 0
	[pt1, pt2, pt3, pt4]
end

#Compute the quads of a Bounding Box
def G6.bbox_quads_lines(bbox)
	corners = []
	for i in 0..7
		corners[i] = bbox.corner i
	end
	if corners[4] == corners[0]
		lq = [0, 1, 3, 2]		
		ll = [0, 1, 1, 3, 3, 2, 2, 0]
	else
		lq = [0, 1, 3, 2, 0, 1, 5, 4, 1, 3, 7, 5, 3, 2, 6, 7, 2, 0, 4, 6, 4, 5, 7, 6]
		ll = [0, 1, 1, 3, 3, 2, 2, 0, 0, 4, 1, 5, 2, 6, 3, 7, 4, 5, 5, 7, 7, 6, 6, 4]
	end	
	[lq.collect { |i| corners[i] }, ll.collect { |i| corners[i] }]
end

#--------------------------------------------------------------------------------
# UV Management
#--------------------------------------------------------------------------------

#Get the real UV at a given 3d Point
def G6.uv_at_point(face, pt, recto, tw=nil)
	unless tw
		@tw = Sketchup.create_texture_writer unless @tw
		tw = @tw
	end	
	uvh = face.get_UVHelper(true, true, tw)
	uv = (recto) ? uvh.get_front_UVQ(pt) : uvh.get_back_UVQ(pt)
	Geom::Point3d.new uv.x / uv.z, uv.y / uv.z, 1
end

#Get the real UV at a given list of 3d Point
def G6.uv_at_points(face, lpt, recto, tw=nil)
	unless tw
		@tw = Sketchup.create_texture_writer unless @tw
		tw = @tw
	end	
	uvh = face.get_UVHelper(true, true, tw)
	luv = []
	if recto
		lpt.each { |pt| luv.push uvh.get_front_UVQ(pt) }
	else
		lpt.each { |pt| luv.push uvh.get_back_UVQ(pt) }
	end
	luv.collect { |uv| Geom::Point3d.new(uv.x / uv.z, uv.y / uv.z, 1) }
end

#Reverse a face, taking into account the possible texturing
def G6.reverse_face_with_uv(face, tw=nil)
	mat = face.material
	bkmat = face.back_material
	
	#No texture on the face
	return face.reverse! unless (mat && mat.texture) || (bkmat && bkmat.texture)
	
	#Face has at a last one side with texture
	unless tw
		@tw = Sketchup.create_texture_writer unless @tw
		tw = @tw
	end	
	uvh = face.get_UVHelper(true, true, tw)

	#Getting the best vertices for the face
	lvx = face.vertices
	n = [4, face.vertices.length].min
	lpt = lvx[0..n-1].collect { |vx| vx.position }
	
	#Getting the texture UV for the front and back sides
	if mat && mat.texture
		lpos_uv_front = []
		lpt.each do |pt|
			uv = uvh.get_front_UVQ(pt)
			ptuv = Geom::Point3d.new(uv.x / uv.z, uv.y / uv.z, 1)
			lpos_uv_front.push pt, ptuv
		end
	end
	if bkmat && bkmat.texture
		lpos_uv_back = []
		lpt.each do |pt|
			uv = uvh.get_back_UVQ(pt)
			ptuv = Geom::Point3d.new(uv.x / uv.z, uv.y / uv.z, 1)
			lpos_uv_back.push pt, ptuv
		end
	end
	
	#Reversing the face
	face.reverse!
	
	#Transfering the texture UV
	face.position_material mat, lpos_uv_front, true if mat && mat.texture
	face.position_material bkmat, lpos_uv_back, true if bkmat && bkmat.texture	
end

#---------------------------------------------------------------------------------------------------------------------------------
# Sketchup Capability
#      - M1 = 8.0.4810+
#	- M2 = 8.0.11751+
#---------------------------------------------------------------------------------------------------------------------------------

def G6.suversion
	Sketchup.version =~ /(.+)\.(.+)\.(.+)/
	[$1.to_i, $2.to_i, $3.to_i]
end

def G6.suversion_after_8M1
	if @suafter_8M1 == nil
		major, minor, build = G6.suversion
		@suafter_8M1 = (major > 8 || (major == 8 && minor >= 0 && build >= 4810))
	end	
	@suafter_8M1
end	

def G6.suversion_after_8M2
	if @suafter_8M2 == nil
		major, minor, build = G6.suversion
		@suafter_8M2 = (major > 8 || (major == 8 && minor >= 0 && build >= 11751))
	end	
	@suafter_8M2
end	

def G6.suversion_after_71
	if @suafter_71 == nil
		major, minor, build = G6.suversion
		@suafter_71 = (major > 7 || (major == 7 && minor >= 1))
	end	
	@suafter_71
end	

#Check if the current version of SU can display polygons with color (>= SU 8.0 M1)
def G6.su_capa_color_polygon ; G6.suversion_after_8M1 ; end
def G6.su_capa_refresh_view ; G6.suversion_after_71 ; end

#---------------------------------------------------------------------------------------------------------------------------------
# Drawing utilities
#---------------------------------------------------------------------------------------------------------------------------------

def G6.draw_face(view, face, color, t)
	return unless face && color
	mesh = face.mesh
	pts = mesh.points
	triangles = []
	mesh.polygons.each do |p|
		triangles += p.collect { |i| t * pts[i.abs-1] }
	end
	view.drawing_color = color
	view.draw GL_TRIANGLES, triangles
end

def G6.face_triangles(face, t)
	return [] unless face
	mesh = face.mesh
	pts = mesh.points
	triangles = []
	mesh.polygons.each do |p|
		triangles += p.collect { |i| t * pts[i.abs-1] }
	end
	triangles
end

def G6.face_polygons(face, t)
	return [] unless face
	mesh = face.mesh
	pts = mesh.points
	polygons = []
	mesh.polygons.each do |p|
		polygons.push p.collect { |i| t * pts[i.abs-1] }
	end
	polygons
end

#---------------------------------------------------------------------------------------------------------------------------------
# Sun Shadow
#---------------------------------------------------------------------------------------------------------------------------------

@@shadow_props = []

#Set or restore the option for shadow setting
#this is needed in SU 7 to make surfaces black when drawn because of a bug in the SU API
def G6.set_sun(shad=nil)
	return unless SU_MAJOR_VERSION >= 7 && Sketchup.version < '8.0.4811'
	shinfo = Sketchup.active_model.shadow_info
	props = ['UseSunForAllShading', 'Light', 'Dark', 'DisplayShadows']
	if @@shadow_props.empty?
		@@shadow_props = props.collect { |a| shinfo[a] }
	end	
	if shad
		shinfo['UseSunForAllShading'] = true
		shinfo['Light'] = 0
		shinfo['Dark'] = 100
		shinfo['DisplayShadows'] = false
	else
		props.each_with_index { |a, i| shinfo[a] = @@shadow_props[i] }
		@@shadow_props = []
	end	
end

@@shadow_toggled_once = nil

#Check if Shadow Switch on / off has been done once - This is only for SU7 and because of a bug in the SU API concerning surface colors
def G6.shadow_toggled_once?
	return true if SU_MAJOR_VERSION < 7 || @@shadow_toggled_once
	@@shadow_toggled_once = true
	false
end

#---------------------------------------------------------------------------------------------------------------------------------
# Curl - Edge chaining methods
#---------------------------------------------------------------------------------------------------------------------------------

#Compute the total length of a curve
def G6.curl_length(pts)
	d = 0.cm
	for i in 1..pts.length-1
		d += pts[i-1].distance pts[i]
	end
	d
end

#Evaluate a range specification as a range
def G6.range(range, ibeg, iend)
	if range == nil
		range = ibeg..iend
	elsif range.class == Fixnum
		range = (range < ibeg && ipos > iend) ? ibeg..iend : range..range
	end
	range
end

#Compute the triplet a points around one or at each points of the curve (based on ipos)
#  --> return a list of points triplet [pt1, pt2, pt3] where
#          - pt2 is the point of the curve
#          - pt1 is the previous point which is different from pt2 or nil
#          - pt3 is the next point which is different from pt2 or nil
def G6.curl_triplets(crvpts, range=nil)
	nb = crvpts.length
	ncrv = nb - 1
	loop = (crvpts.first == crvpts.last)

	ltriplets = []
	range = G6.range range, 0, ncrv
	
	for i in range
		pt1 = pt3 = nil
		pt2 = crvpts[i]
		for j in 1..ncrv
			k = i + j
			k -= ncrv if k > ncrv && loop
			break if k > ncrv
			pt3 = crvpts[k]
			break unless pt3 && pt3 == pt2
		end	
		for j in 1..ncrv
			k = i - j
			k += ncrv if k < 0 && loop
			break if k < 0
			pt1 = crvpts[k]
			break unless pt1 && pt1 == pt2
		end	
		pt1 = nil if pt1 == pt2
		ltriplets[i] = [pt1, pt2, pt3]
	end
	ltriplets
end

#Compute the mid point of a curl
def G6.curl_mid_point(pts)
	n = pts.length - 1
	return pts[0] if n == 0
	return Geom.linear_combination(0.5, pts[0], 0.5, pts[1]) if n == 1
	d = 0
	tdist = [0]
	for i in 1..n
		d += pts[i].distance(pts[i-1])
		tdist.push d
	end
	mid_dist = tdist.last * 0.5
	return pts[0] if mid_dist == 0
	
	ibeg = 1
	for i in 1..n
		if tdist[i] >= mid_dist 
			ratio = (tdist[i] - mid_dist) / (tdist[i] - tdist[i-1])
			return Geom.linear_combination(ratio, pts[i-1], 1 - ratio, pts[i])
		end	
	end
	pts[n/2]
end

#Sample points on a curve based on a division by a number n
def G6.curl_sample(pts, n)
	dtot = G6.curl_length pts
	delta = dtot / n
	
	lspt = [pts[0]]
	d = 0
	dcur = delta
	
	for i in 0..pts.length-2
		dseg = pts[i].distance pts[i+1]
		d += dseg
		while true
			break if d <= dcur
			r = (d - dcur) / dseg
			lspt.push Geom.linear_combination(r, pts[i], 1-r, pts[i+1])
			dcur += delta
		end	
	end
	lspt.push pts.last
	lspt
end

#Determine if two open curves must be joined by beg-beg or beg-end
#   Return 1 if beg-beg
#   Return -1 if beg-end
def G6.curl_trigo_sense(pts1, pts2)
	n = 10
	spt1 = G6.curl_sample pts1, n
	spt2 = G6.curl_sample pts2, n
	spt2r = spt2.reverse
	
	dsum = dsum_rev = 0
	for i in 0..n
		dsum += spt1[i].distance(spt2[i])
		dsum_rev += spt1[i].distance(spt2r[i])
	end
	
	(dsum <= dsum_rev) ? 1 : -1
end

#Check if a sequence of points is aligned
def G6.curl_is_aligned?(pts)
	n = pts.length - 1
	return true if n < 2
	vecprev = nil
	for i in 1..n
		vec = pts[i-1].vector_to pts[i]
		next unless vec.valid?
		if vecprev
			return false unless vec.parallel?(vecprev)
		end	
		vecprev = vec
	end
	true
end

#Get theTOS attribute for an edge - nil if not part of a TOS curve
def G6.curl_tos(edge)
	edge.get_attribute 'skp', 'ToolOnSurface'
end

#Compute the list of edges belonging to a curl  or a curve from a given edge
def G6.curl_edges(edge)
	tos_attr = G6.curl_tos(edge)
	if tos_attr
		hsh = { edge.entityID => edge }
		ls = G6.curl_tos_extend edge, tos_attr, hsh
		while !ls.empty?
			ll = []
			ls.each { |e| ll += G6.curl_tos_extend(e, tos_attr, hsh) }
			ls = ll
		end
		return hsh.values	
	end
	curve = edge.curve
	return [edge] unless curve
	curve.edges
end

#Compute the list of edges belonging to a curl from a given edge
def G6.curl_tos_extend(edge, tos_attr, hsh_edges)
	ls = []
	[edge.start, edge.end].each do |vertex|
		le = vertex.edges.find_all { |e| !hsh_edges[e.entityID] &&  G6.curl_tos(e) == tos_attr }
		if le.length == 1
			e1 = le[0]
			hsh_edges[e1.entityID] = e1
			ls.push e1
		end	
	end
	ls
end

#Compute the edges connected to an edge.
#This function is required because 'all_connected' method has a side effect to deactivate the visibility of the selection
def G6.curl_all_connected(edge)
	hsh_v = {}
	hsh_e = {}
	vstart = edge.start
	all_v = [vstart]
	lst_v = [vstart]
	hsh_v[vstart.entityID] = vstart
	while true
		new_v = []
		lst_v.each do |vertex|
			vertex.edges.each do |e|
				v = e.other_vertex vertex
				unless hsh_v[v.entityID]
					new_v.push v 
					hsh_v[v.entityID] = v
				end	
			end
		end	
		break if new_v.length == 0
		lst_v = new_v.clone
		all_v += new_v
	end
	
	all_v.each do |v|
		v.edges.each do |e|
			hsh_e[e.entityID] = e
		end
	end	
	hsh_e.values
end

#extend selection in follow mode for edge at vertex
def G6.curl_follow_extend(edge, anglemax, stop_at_crossing=false, &proc)
	common_normal = nil
	ls_edges = []
	[edge.start, edge.end].each do |vertex|
		edgenext = edge
		while edgenext
			le = vertex.edges.to_a.find_all { |ee| ee != edgenext && G6.edge_plain?(ee) }
			len = le.length
			if len == 1
				e = le[0]
				an = Math::PI - G6.edges_angle_at_vertex(edgenext, e, vertex)
				#ls = (an > anglemax) ? [] : [[an, e]]
				ls = [[an, e]]
			elsif len > 0 && stop_at_crossing
				break
			else	
				le = vertex.edges.to_a.find_all { |ee| ee != edgenext } if len == 0 ####
				ls = []
				le.each do |e|
					next if e == edgenext
					next if proc && !proc.call(e)
					an = Math::PI - G6.edges_angle_at_vertex(edgenext, e, vertex)
					next if an > anglemax
					if common_normal && common_normal.valid?
						vn = G6.edges_normal(e, edgenext)
						if vn.valid? && vn.parallel?(common_normal)
							ls.push [an, e] 
							next
						end	
					end	
					if e.common_face(edgenext)
						ls.push [an, e]
					elsif an < anglemax
						ls.push [an, e]
					end	
				end
			end
			break if ls.length == 0
			
			ls.sort! { |a1, a2| a1[0] <=> a2[0] } if ls.length > 1
			e = ls[0][1]
			break if e == edge || ls_edges.include?(e)
			common_normal = G6.edges_normal(e, edgenext) unless common_normal && common_normal.valid?
			edgenext = e
			vertex = e.other_vertex(vertex)
			ls_edges.push edgenext
			
		end	#while true
	end	#loop on end and start
	
	return ls_edges
end

#Remove the duplicated points (except if at beginning and end)
def G6.curl_deduplicate(crvpts)
	new_pts = []
	for i in 0..crvpts.length-1
		pt = crvpts[i]
		new_pts.push pt if pt != new_pts.last
	end
	new_pts
end

#Concatenate a list of sequences of points
def G6.curl_concat(lpt)
	lres = []
	lpt.each do |pts|
		lres += (pts[0] == lres.last && pts.length > 1) ? pts[1..-1] : pts
	end	
	lres	
end

#Harmonize two curves to have the same number of matching vertices
def G6.curl_harmonize(curl1, curl2, simplify_factor=0.025)
	return [curl1, curl2] if curl1 == nil || curl2 == nil

	#Special case when one curve is reduced to a single point
	if curl1.length == 1
		curl1 = Array.new curl2.length, curl1.first
		return [curl1, curl2]
	end
	if curl2.length == 1
		curl2 = Array.new curl1.length, curl2.first
		return [curl1, curl2]
	end
	
	#computing the total distance for first curve
	n1 = curl1.length
	dtot1 = 0.0
	lnorm1 = [[0.0, 0, nil]]
	for i in 1..n1-1
		dtot1 += curl1[i].distance(curl1[i-1])
		lnorm1.push [dtot1, i, nil]
	end	
	
	#computing the total distance for second curve
	n2 = curl2.length
	dtot2 = 0.0 
	lnorm2 = [[0.0, nil, 0]]
	for i in 1..n2-1
		dtot2 += curl2[i].distance(curl2[i-1])
		lnorm2.push [dtot2, nil, i]
	end	
	
	#Matching the curls
	ratio = dtot2 / dtot1
	lnorm1.each { |ll| ll[0] = ll[0] / dtot1 }
	lnorm2.each { |ll| ll[0] = ll[0] / dtot2 }

	lnorm = (lnorm1 + lnorm2).sort { |a, b| a[0] <=> b[0] }
	
	i1 = 0
	i2 = 0
	lpairs = [[curl1[0], curl2[0], 0, 0]]
	lnorm[2..-3].each do |ll|
		d = ll[0]
		j1 = ll[1]
		j2 = ll[2]
		if j1
			i1 = j1
			pt1 = curl1[i1]
			if lnorm2[i2][0] == d
				pt2 = curl2[i2]
				j2 = i2
			elsif d == 1.0
				pt2 = curl2.last
				j2 = i2 + 1
			else	
				ratio = (d - lnorm2[i2][0]) / (lnorm2[i2+1][0] - lnorm2[i2][0])
				pt2 = Geom.linear_combination 1.0 - ratio, curl2[i2], ratio, curl2[i2+1]
			end	
		elsif j2
			i2 = j2
			pt2 = curl2[i2]
			if lnorm1[i1][0] == d
				pt1 = curl1[i1]
				j1 = i1
			elsif d == 1.0
				pt1 = curl1.last
				j1 = i1 + 1
			else	
				ratio = (d - lnorm1[i1][0]) / (lnorm1[i1+1][0] - lnorm1[i1][0])
				pt1 = Geom.linear_combination 1.0 - ratio, curl1[i1], ratio, curl1[i1+1]
			end	
		end	
		lpairs.push [pt1, pt2, j1, j2]
	end
	lpairs.push [curl1[-1], curl2[-1], n1-1, n2-1]
	
	#Removing duplicates
	lpairs_final = [lpairs[0]]
	n = lpairs.length
	for i in 1..n-1
		pair = lpairs[i]
		pair_prev = lpairs_final.last
		if pair[0].distance(pair_prev[0]) == 0.0 || pair[1].distance(pair_prev[1]) == 0.0
			pair_prev[2] = pair_prev[3] = :node
		else
			lpairs_final.push pair
		end	
	end	

	#Scanning the pairs to check which one should be merged
	lmerge = []
	lp = lpairs_final
	n = lp.length - 1
	for i in 1..n
		pair_prev = lp[i-1]
		pair = lp[i]
		d1 = pair[0].distance(pair_prev[0]) / dtot1
		d2 = pair[1].distance(pair_prev[1]) / dtot2
		
		if d1 <= simplify_factor && d2 <= simplify_factor
			if pair[2] && !pair[3] && pair_prev[3] && !pair_prev[2]
			   lmerge.push [i, i-1, d1, d2]
			elsif !pair[2] && pair[3] && !pair_prev[3] && pair_prev[2]
			   lmerge.push [i-1, i, d1, d2]
			 end
		end
	end

	#Filtering the merges
	lmerge.sort! { |a, b| (a[2] + a[3]) <=> (b[2] + b[3]) }
	
	hexclude = {}
	htreated = {}
	lmerge.each do |lm|
		i1 = lm[0]
		i2 = lm[1]
		pair1 = lp[i1]
		pair2 = lp[i2]
		next if htreated[i1] || htreated[i2]
		d1 = lm[2]
		d2 = lm[3]
		if d1 < d2
			pair1[3] = :node
			pair1[1] = pair2[1]
			hexclude[i2] = true
		else
			pair2[2] = :node
			pair2[0] = pair1[0]
			hexclude[i1] = true
		end
		htreated[i1] = htreated[i2] = true
	end
	
	#Executing the merge
	lpairs_final = []
	for i in 0..lp.length-1
		lpairs_final.push lp[i] unless hexclude[i]
	end	
	
	#Rebalancing the pairs
	lp = lpairs_final
	n = lp.length - 2
	for i in 1..n
		pair_prev = lp[i-1]
		pair_next = lp[i+1]
		pair = lp[i]
		if pair[2] && pair[3] == nil
			d1 = pair_next[0].distance(pair_prev[0])
			d = pair[0].distance(pair_prev[0])
			ratio = d / d1
			pair[1] = Geom.linear_combination 1.0 - ratio, pair_prev[1], ratio, pair_next[1]
		
		elsif pair[2] == nil && pair[3]
			d2 = pair_next[1].distance(pair_prev[1])
			d = pair[1].distance(pair_prev[1])
			ratio = d / d2
			pair[0] = Geom.linear_combination 1.0 - ratio, pair_prev[0], ratio, pair_next[0]		
		end
	end
	
	#Returning the new curl value
	curl1 = lpairs_final.collect { |pair| pair[0] }
	curl2 = lpairs_final.collect { |pair| pair[1] }

	[curl1, curl2]	
end

#----------------------------------------------------------------------------
# Utilities for Curl Deformations
#----------------------------------------------------------------------------

#Compute the avreage of twin curves (assumed to have matching points)
def G6.curl_average_curve(pts1, pts2, smooth=false)
	n = pts1.length - 1
	d1 =[0.0]
	d2 =[0.0]
	iend = nil
	ibeg = nil
	for i in 1..n
		d1[i] = d1[i-1] + pts1[i-1].distance(pts1[i])
		ibeg = i if d1[i] == d1[i-1]
		d2[i] = d2[i-1] + pts2[i-1].distance(pts2[i])
		if iend == nil && d2[i] == d2[i-1]
			iend = i-1
		end	
	end	
	dtot1 = d1.last
	dtot2 = d2.last
	
	pts_res = [pts1[0]]
	for i in 1..n
		if d1[i] == 0
			ratio = 0.0
		elsif d2[i] == 0 || d2[i] == dtot2
			ratio = 1.0
		else	
			ratio1 = d1[i] / dtot1
			ratio2 = d2[i] / dtot2
			a = ratio2
			b = ratio1
			ratio = (a * ratio1 + b * ratio2) / (a + b)
			ratio = ratio1 / (1.0 + ratio1 - ratio2)
		end	
		puts "ratio too high #{ratio}" if ratio > 1
		ratio = G6.scurve_ratio ratio if smooth
		pts_res.push Geom.linear_combination(1 - ratio, pts1[i], ratio, pts2[i])
	end	
	pts_res
end

#Compute a smooth ratio between 0 and 1 (method Cubic Bezier)
def G6.scurve_ratio(t)
	3 * t * t * (1 - t) + t * t * t
end

#----------------------------------------------------------------------
# Curve Stretching
#----------------------------------------------------------------------

#Deform a sequence of points from the translation of the last point
def G6.curl_move_extremities(pts, ptbeg, ptend)
	if ptbeg == ptend
		return Array.new(pts.length, ptbeg)
	end	

	vec = pts[0].vector_to ptbeg
	if vec.valid?
		t = Geom::Transformation.translation vec
		pts = pts.collect { |pt| t * pt }
	end	
	G6.curl_move_end_point pts, ptend
end

#Deform a sequence of points from the translation of the last point
def G6.curl_move_end_point(pts, ptarget)
	ptend = pts.last
	ptbeg = pts.first
	return pts if ptend == ptarget
	return pts if ptbeg == ptarget
	
	if pts[0] == pts[1]
		return Array.new(pts.length, ptarget)
	end	
	for i in 1..pts.length-1
		puts "DOUBLE PTS = #{i} - pt = #{pts[i]}" if pts[i-1] == pts[i]
	end
	
	#Computing the base and target configuration
	vec_base = ptbeg.vector_to ptend
	vec_targ = ptbeg.vector_to ptarget
	normal = vec_base * vec_targ
	if normal.valid?
		angle = vec_base.normalize.angle_between vec_targ.normalize
		t = Geom::Transformation.rotation ptbeg, normal, angle
	else
		t = Geom::Transformation.new
	end	
	dbase = ptbeg.distance(ptend)
	dtarg = ptbeg.distance(ptarget)
	ratio = dtarg / dbase
	
	#Rotating the curl
	pts_rot = pts.collect { |pt| t * pt }
	
	#Constructing the resulting curve by stretching the segments
	pts_final = [ptbeg]
	for i in 1..pts_rot.length-1
		pt1 = pts[i-1]
		pt2 = pts[i]
		d = pt1.distance(pt2) * ratio
		puts "PT DOUBLE = #{i} - #{pts_rot[i-1]}" if pts_rot[i-1] == pts_rot[i]
		vec = pts_rot[i-1].vector_to pts_rot[i]
		pt = (vec.valid?) ? pts_final.last.offset(vec, d) : pts_final.last	
		pts_final.push pt
	end
	pts_final
end

#----------------------------------------------------------------------
# Area Calculations
#----------------------------------------------------------------------

#Area of a planar polygon (points given in 2D)
def G6.polygon_area(lpt_2d)
	area = 0.0
	for i in 0..lpt_2d.length-2
		pt1 = lpt_2d[i]
		pt2 = lpt_2d[i+1]
		area += pt1.x * pt2.y - pt2.x * pt1.y
	end
	0.5 * area.abs
end

#Calculate the area of a face, possibly with holes
def G6.face_area(face, t=nil)
	if t
		loops = face.loops.collect { |loop| loop.vertices.collect { |v| t * v.position } }
		pt1 = face.outer_loop.vertices[0].position
		normal = (t * pt1).vector_to(t * pt1.offset(face.normal, 1))
	else
		loops = face.loops.collect { |loop| loop.vertices.collect { |v| v.position } }
		pt1 = face.outer_loop.vertices[0].position
		normal = pt1.vector_to(pt1.offset(face.normal, 1))
	end

	#Finding the plane
	axes = normal.axes
	tr_axe = Geom::Transformation.axes loops[0][0], *axes
	tr_axe_inv = tr_axe.inverse
	
	#Area for outer_loop
	lpt_2d = loops[0].collect { |pt| tr_axe_inv * pt }
	lpt_2d.push lpt_2d[0]
	area = G6.polygon_area lpt_2d
	
	#Area for inner loops
	if loops.length > 1
		loops[1..-1].each do |loop|
			lpt_2d = loop.collect { |pt| tr_axe_inv * pt } 
			lpt_2d.push lpt_2d[0]
			area -= G6.polygon_area lpt_2d
		end
	end	
	area
end

#---------------------------------------------------------------------------------------------
# Utilities for picking components
#---------------------------------------------------------------------------------------------

#Find the closest edge to the mouse
def G6.closest_entity(view, ip, x, y, &check_edge_proc)
	ip_entity = [ip.vertex, ip.edge, ip.face].find { |e| e }
	return ip_entity if !ip_entity.instance_of?(Sketchup::Vertex)
	ledges = ip_entity.edges
	ledges = ledges.find_all { |e| check_edge_proc.call(e) } if check_edge_proc
	return ip.face if ledges.empty?
	return ledges[0] if ledges.length == 1
	
	#At vertex: finding the closest edge
	ls = []
	lsback = []
	tr = ip.transformation
	ptvx = tr * ip.vertex.position
	ptxy = Geom::Point3d.new x, y
	ledges.each do |edge|
		ptbeg = view.screen_coords(tr * edge.start.position)
		ptend = view.screen_coords(tr * edge.end.position)
		ptproj = ptxy.project_to_line([ptbeg, ptend])
		if G6.point_within_segment?(ptproj, ptbeg, ptend)
			ls.push [edge, ptxy.distance(ptproj)]
		else	
			lsback.push [edge, ptproj.distance(ptvx)]
		end	
	end
	if ls.length > 0
		ls.sort! { |a, b| a[1] <=> b[1] }
		return ls.first[0]
	end
	lsback.sort! { |a, b| a[1] <=> b[1] }
	lsback.first[0]
end

#Determining the component parent and the transformation
@@tr_id = Geom::Transformation.new
@@origin2 = Geom::Point3d.new 1, 0, 0

def G6.pick_component(view, ip, x, y, entity=nil)
	#Getting the entity, if not provided
	ip.pick view, x, y
	entity_ip = G6.closest_entity(view, ip, x, y)
	entity = entity_ip unless entity
	
	#Getting the hierachical list of groups and components
	model = Sketchup.active_model
	ll = model.raytest view.pickray(x, y)
	parent = model
	tr = @@tr_id
	if ll	
		lcomp = ll[1].reverse.find_all { |e| G6.is_grouponent?(e) }
		comp = lcomp.first
		if comp	
			lcomp.each { |c| tr = c.transformation * tr }
			parent = comp
		end	
	end
	if entity == nil || entity.parent == model || entity.parent == nil
		parent = model
		tr = @@tr_id
	elsif entity.parent != G6.grouponent_definition(parent) 
		if entity.parent.instances.length >= 1
			parent = entity.parent.instances[0]
			ph = view.pick_helper
			ph.do_pick x, y
			tr = ph.transformation_at 0
		else
			entity = nil
		end	
	end
	parent = model if model.active_entities == G6.entities_from_parent(parent)
	if tr == nil || (parent != model && tr.identity?)
		parent = model
		tr = (ip.transformation.identity?) ? @@tr_id : ip.transformation
	end	
	
	[parent, tr]
end

def G6.parent_transformation(parent)
	tr = @@tr_id
	return tr
	while true
		parent = parent.parent
		break if parent == Sketchup.active_model
		tr = parent.transformation * tr
	end
	tr
end

#Retrieve the entities from parent
def G6.entities_from_parent(parent)
	if parent.instance_of?(Sketchup::ComponentInstance)
		entities = parent.definition.entities
	elsif parent.instance_of?(Sketchup::Group)
		entities = parent.entities
	else
		entities = Sketchup.active_model.active_entities
	end
	entities	
end

#Draw the component contour if any
def G6.draw_component(view, parent, tr, default=false)
	return unless parent && parent.valid? && parent != Sketchup.active_model
	if default
		view.drawing_color = 'gray'
		if parent.instance_of?(Sketchup::ComponentInstance)
			view.line_stipple = ''
			view.line_width = 1
		else
			view.line_stipple = '-'
			view.line_width = 2
		end
	end	
	llines = G6.grouponent_box_lines(view, parent, tr)
	view.draw GL_LINES, llines unless llines.empty?
end

#Compute a losange label with lead line at position ipos of a 3d point array
#Return [pt_losange, pt1, pt2] information in 2D for drawing
def G6.compute_losange_label(view, pts, ipos=0, ratio=1, length=10)
	n = pts.length - 1
	ipos = 0 if ipos == nil || n == 0
	pt1 = pts[ipos]
	pt2 = pts[ipos-1]
	pt2 = pt1 unless pt2
	ratio = 1 if ratio == nil || ratio < 0 || ratio > 1
	ptmid = Geom.linear_combination ratio, pt1, 1-ratio, pt2
	pt2d = view.screen_coords ptmid
	pt1_2d = view.screen_coords pt1
	pt2_2d = view.screen_coords pt2
	
	#Computing the lead line
	if ratio > 0 || ipos < 3 || ipos == n
		vec = pt1_2d.vector_to pt2_2d
		vec2d = (vec.valid?) ? vec * Z_AXIS : Y_AXIS
	else	
		vec = Geom.linear_combination 0.5, pt1.vector_to(pts[ipos-1]).normalize, 0.5, pt1.vector_to(pt2).normalize
		vec2d = (vec.valid?) ? pt2d.vector_to(@view.screen_coords(pt1.offset(vec, 10))) : Y_AXIS
	end
	vec2dp = vec2d * Z_AXIS

	length = 10 unless length
	dec = length + 3
	ptend = pt2d.offset vec2d, -length
	if ptend.y < dec
		ptend = pt2d.offset vec2d, length
		dec = -dec
	end	
		
	#Compute the losange
	dec2 = dec / 2
	ptop = ptend.offset vec2d, -dec
	ptmid = ptend.offset vec2d, -dec2
	pt1 = ptmid.offset vec2dp, dec2
	pt2 = ptmid.offset vec2dp, -dec2
	pts_losange = [ptend, pt1, ptop, pt2]
	pts_losange.each { |pt| pt.z = 0 }
	
	[[pt2d, ptend], pts_losange]
end

#Draw a losange label
def G6.draw_losange_label(view, lead_line, losange, color)
	#Draw the leading line
	view.drawing_color = color
	view.line_width = 1
	view.line_stipple = '-'
	view.draw2d GL_LINE_STRIP, lead_line
	
	#Draw the losange
	pts = losange
	view.draw2d GL_QUADS, pts 
	view.line_width = 1
	view.line_stipple = ''
	view.drawing_color = 'gray'
	view.draw2d GL_LINE_LOOP, pts
end

#---------------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------------
# Mouse Picking utilities (based on PickHelper)
#---------------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------------

#Determine Face, Edge and Element under the mouse
def G6.picking_under_mouse(x, y, view, precision=nil)
	ph = view.pick_helper
	(precision.class == Fixnum) ? ph.do_pick(x, y, precision) : ph.do_pick(x, y)
	
	ph_edge = ph.picked_edge
	ph_face = ph.picked_face
	ph_elt = ph.picked_element
	lst_elt = [ph_face, ph_edge, ph_elt]
				
	#Finding the parent and transformation
	lst_info = []
	lst_elt.each do |elt|
		unless elt
			lst_info.push nil
			next
		end	
		tr = nil
		parent = nil
		for i in 0..ph.count
			ls = ph.path_at(i)
			if ls && ls.include?(elt)
				parent = ls[-2]
				tr = ph.transformation_at(i)
				break
			end
		end	
		tr = @@tr_id unless tr
		lst_info.push [elt, tr, parent]
	end	
	lst_info
end

#Determine Edge under the mouse
def G6.picking_edge_under_mouse(x, y, view, precision=nil)
	ph = view.pick_helper
	(precision.class == Fixnum) ? ph.do_pick(x, y, precision) : ph.do_pick(x, y)
	
	ph_edge = ph.picked_edge
	return nil unless ph_edge
	
	#Finding the parent and transformation
	tr = nil
	parent = nil
	for i in 0..ph.count
		ls = ph.path_at(i)
		if ls && ls.include?(ph_edge)
			parent = ls[-2]
			tr = ph.transformation_at(i)
			break
		end
	end	
	[ph_edge, tr, parent]
end

#---------------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------------
# CAMERA: Camera management
#---------------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------------

def G6.camera_duplicate(camera)
	if camera.perspective?
		camera_new = Sketchup::Camera.new camera.eye, camera.target, camera.yaxis, camera.perspective?, camera.fov
	else	
		camera_new = Sketchup::Camera.new camera.eye, camera.target, camera.yaxis, camera.perspective?
	end	
	camera_new.description = camera.description
	camera_new
end

#---------------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------------
# ViewTracker: track if view has changed
#---------------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------------

class ViewTracker

def initialize
	@view = Sketchup.active_model.active_view
	@ptref = [ORIGIN, ORIGIN.offset(X_AXIS), ORIGIN.offset(Y_AXIS)]
	@ptref_2d = nil
end

def changed?(reference=nil)
	reference = @ptref_2d unless reference
	view = Sketchup.active_model.active_view
	ptref_2d = @ptref.collect { |pt| view.screen_coords pt }
	if !@ptref_2d || view != @view || ptref_2d[0] != reference[0] || ptref_2d[1] != reference[1] || ptref_2d[2] != reference[2]
		@view == view
		@ptref_2d = ptref_2d
		return @ptref_2d
	end
	false
end
	
def reference
	@ptref.collect { |pt| Sketchup.active_model.active_view.screen_coords pt }
end
	
end	#class ViewTracker

#---------------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------------
# SPACEGRID: Create spatial grids
#---------------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------------

class SpaceGrid

#SPACE_GRID: create the instance
def initialize(bbtot, nspace, tol=nil)
	#Parameters for the spatial grid
	@nv = []
	@delta = []
	@vmin = []
	@tol = tol
	@hsh_bb = {}
	
	#Parameters of the grid. If tolerance is given, then the grid is spaced by the tolerance, not the specified <nspace>
	[[0, 1], [1, 2], [0, 4]].each_with_index do |a, k|
		i, j = a
		dtot = bbtot.corner(i).distance bbtot.corner(j)
		n = (tol) ? (dtot / tol).ceil : nspace
		@nv[k] = (dtot > 0) ? n : 1
		@delta[k] = dtot / @nv[k]
		@vmin[k] = bbtot.corner(i)[k]
	end	
end

#SPACE_GRID: Get the Box hash array
def get_hsh_bb ; @hsh_bb ; end

#SPACEGRID: Classify a set of points / vertices into a 3D grid	
def where_is_point(pt)
	keys = [[], [], []]
	for k in 0..2
		v = pt[k] - @vmin[0]
		dv = @delta[k]
		n = @nv[k]
		iv = (dv == 0) ? 0 : (v / dv).floor
		keys[k].push iv
		next unless @tol
		keys[k].push iv-1 if iv > 0
		keys[k].push iv+1 if iv < n
	end	
	lkeys = []
	keys[0].each do |k0|
		keys[1].each do |k1|
			keys[2].each do |k2|
				lkeys.push "#{k0} - #{k1} - #{k2}"
			end
		end
	end
	lkeys
end

#SPACEGRID: Classify a set of points / vertices into a 3D grid	
#Linfo is either a list of 3D points or a list of arrays where the point 3D is the first element
def classify_points_vertices(linfo, &item_proc)		
	linfo.each_with_index do |info, ivx|
		#Notifying the caller
		item_proc.call(ivx) if item_proc
		
		#Classifying the point
		pt = (info.class == Array) ? info.first : info
		lskeys = where_is_point(pt)
		lskeys.each do |key|
			lst = @hsh_bb[key]
			lst = @hsh_bb[key] = [] unless lst
			lst.push info
		end	
	end	
	@hsh_bb
end
	
end	#class SpaceGrid
	
#=============================================================================================
#=============================================================================================
# Class ZoomVoid: Manage zoom in empty space when Sketchup does not do it
#=============================================================================================
#=============================================================================================

class ZoomVoid

#Initialization
def initialize
	@ph = Sketchup.active_model.active_view.pick_helper	
	@ip = Sketchup::InputPoint.new	
end

#Registration at suspend
def suspend(view, x, y)
	@eye_vec = nil
	return unless x && y
	camera = view.camera
	return unless camera.perspective?
	
	#Checking if there is an object under the mouse. If so, let SU do the zoom
	@ph.do_pick x, y
	return if @ph.picked_face || @ph.picked_edge
	elt = @ph.picked_element
	return if elt && !(elt.instance_of?(Sketchup::Drawingelement) && elt.bounds.diagonal == 0)
	@ip.pick view, x, y
	
	#Storing the view configuration
	@eye_point = Geom.intersect_line_plane view.pickray(x, y), [@ip.position, camera.direction]
	@eye_vec = camera.eye.vector_to @eye_point
end

#Performing the zoom when resuming
def resume(view)
	return unless @eye_vec
	camera = view.camera
	eye = camera.eye
	vec = eye.vector_to @eye_point
	
	#Only handle Zoom not Orbit
	return unless vec.parallel?(@eye_vec)
	
	#Zoom already handled by Sketchup
	ratio = 0.1
	d0 = @eye_vec.length
	d = eye.distance @eye_point
	return if (d - d0).abs / d0 > ratio
	
	#Compute new camera position depending on zoom in or zoom out	
	offset = (d < d0) ? d : -d
	new_eye = eye.offset vec, offset * ratio
	
	#Computing the position for target
	dir = camera.direction
	new_target = Geom.intersect_line_plane [new_eye, dir], [camera.target, dir]	
	camera.set new_eye, new_target, camera.up
end
		
end	#class ZoomVoid
	
end #Module G6

